package org.openmuc.core.driver;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.IOException;
import java.util.List;

import org.openmuc.core.measurement.ChannelIdentifier;
import org.openmuc.core.measurement.Measurement;

/**
 * 
 * Interface that all openMUC i/o drivers have to implement. Drivers that
 * implement this interface can be plugged into the framework
 * 
 * 
 * @author mzillgit
 * 
 */
public interface DeviceDriverInterface {
	public static final int ERROR_TIMEOUT = 1;
	public static final int ERROR_INVALID_MESSAGE = 2;
	public static final int IO_ERROR = 3;

	/************************************************************************************
	 * Driver and device configuration operations
	 * 
	 ***********************************************************************************/

	/**
	 * Configure the interface.
	 * 
	 * @param iface
	 *            Interface descriptor. E.g. com1 for "/dev/ttyS0" usb1 for
	 *            "/dev/ttyUSB0"
	 * 
	 * @param configStr
	 *            Configuration data for interface e.g. "setting=2400-7N1"
	 * 
	 *            mbus:com1:p1?setting=2400-7N1
	 * 
	 * @throws NoSuchInterfaceException
	 *             if the interface is not present
	 * @throws MethodNotImplementedException
	 *             if method is not implemented by driver
	 * @throws IOException
	 */
	public void configureInterface(String iface, String configStr) throws NoSuchInterfaceException,
			MethodNotImplementedException, IOException;

	/**
	 * Get a list of supported operations by the driver. This list does not
	 * contain the mandatory operations.
	 * 
	 * @return List of the supported optional operation names
	 * 
	 */
	public List<String> getSupportedOperations();

	public String getDriverID();

	/**
	 * 
	 * Scan the bus for connected meters
	 * 
	 * This operation is optional and my not be supported by specific drivers
	 * 
	 * @param url
	 * @return List of URLs of connected meters or null if this operation is not
	 *         supported by the driver.
	 * @throws MethodNotImplementedException
	 *             if method is not implemented by driver
	 */
	public List<DeviceURI> getDeviceList(BusURI url) throws MethodNotImplementedException, IOException;

	/**
	 * 
	 * Scan a device for available channels
	 * 
	 * This operation is optional and my not be supported by specific drivers
	 * 
	 * @param url
	 * @return List of Strings of assigned channels for a meters or null if this
	 *         operation is not supported by the driver.
	 * @throws MethodNotImplementedException
	 *             if method is not implemented by driver
	 */
	public List<ChannelInformation> getDeviceChannels(DeviceURI url) throws MethodNotImplementedException, IOException;

	/************************************************************************************
	 * Operations for meter data acquisition
	 ***********************************************************************************/

	/**
	 * 
	 * @param url
	 *            URL of the device
	 * 
	 * @param channels
	 *            List of channels to readout. If channels == null all channels
	 *            will be read.
	 * 
	 * @param callback
	 *            Callback object if task finished or error/timeout occured.
	 * @throws MethodNotImplementedException
	 *             if method is not implemented by driver
	 * @throws IOException
	 */
	public void readDeviceAsync(DeviceURI url, List<ChannelIdentifier> channels, DataReceivedListener callback,
			Object handle) throws MethodNotImplementedException;

	/**
	 * 
	 * @param url
	 * @param channels
	 * @throws MethodNotImplementedException
	 *             if method is not implemented by driver
	 * @throws IOException
	 */
	public List<Measurement> readDeviceSync(DeviceURI url, List<ChannelIdentifier> channels)
			throws MethodNotImplementedException, IOException;

	/**
	 * 
	 * Write a channel/register on connected meter / device / i/o
	 * 
	 * @param url
	 *            - URL of the device in a driver specific format
	 * @param channel
	 *            - channel to write to
	 * @param value
	 *            - value to write to register
	 * @throws MethodNotImplementedException
	 *             if method is not implemented by driver
	 */
	public void writeChannel(DeviceURI url, ChannelIdentifier channel, double value)
			throws MethodNotImplementedException, IOException;

}
