package org.openmuc.core.collector;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.util.ArrayList;

import org.apache.log4j.Logger;
import org.openmuc.core.datamanager.DataManager;
import org.openmuc.core.datamanager.config.ConfigurationListener;
import org.openmuc.core.datamanager.config.ConfigurationService;
import org.openmuc.core.driver.DeviceDriverManager;
import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;

public class Activator implements BundleActivator, ConfigurationListener {

	private static Logger logger = Logger.getLogger(Activator.class);
	
	private ConfigurationService configurationService;
	private Configuration configuration;
	private BundleContext context;
	
	private ArrayList<Reader>  dataReaders;
	
	public void start(BundleContext context) throws Exception {
		this.context = context;
		
		startReaders();
	}

	public void stop(BundleContext context) throws Exception {
		stopAllReaders();
	}
	
	private void startReaders() throws Exception {
		/* get the data manager service */
		ServiceReference serviceRef = context.getServiceReference(DataManager.class.getName());
		DataManager dataManager = (DataManager) context.getService(serviceRef);
		
		DeviceDriverManager driverManager;
		
		/* get the meter manager service */
		serviceRef = 
			context.getServiceReference(DeviceDriverManager.class.getName());
		
		driverManager = (DeviceDriverManager) context.getService(serviceRef);
		
		if (driverManager == null) {
			logger.error("No meter driver manager service present");
			throw new Exception("No meter driver manager found");
		}
		
		/* get the configuration service */
		serviceRef = context.getServiceReference(ConfigurationService.class.getName());
		configurationService = (ConfigurationService) context.getService(serviceRef);
		
		configuration = new Configuration(configurationService, dataManager);
				
		/* Start a separate readout process per driver */
		dataReaders = new ArrayList<Reader>();
		int readerCounter = 0;
		
		for (String driver : configuration.getDrivers()) {
			Reader reader = new Reader(driverManager, configuration, 
					configuration.getMeterListForDriver(driver));
			
			dataReaders.add(reader);
			
			Thread thread = new Thread(reader, "mux/reader" + (++readerCounter));
			logger.debug("Start thread " + thread.getName());
			thread.start();
		}	
	}

	private void stopAllReaders() {
		/* Force all readers to stop */
		for (Reader reader : dataReaders) {
			reader.finish();
		}
		
		/* Waiting until all threads are finished */
		while (!haveAllReadersFinished()) {
			try {
				Thread.sleep(10);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}
	
	private boolean haveAllReadersFinished() {
		for (Reader reader : dataReaders) {
			if (reader.isRunning()) return false;
		}
		
		return true;
	}

	public void configurationChanged() {
		stopAllReaders();
		try {
			startReaders();
		} catch (Exception e) {
			logger.error("Failed to start reader threads");
		}
	}
}
