package org.openmuc.core.collector;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * ------------------------------------------------------------------------------
 * 
 * This class manages the readout process for all devices that use the same
 * device driver.
 * 
 */

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.openmuc.core.datamanager.DataStorage;
import org.openmuc.core.datamanager.config.DeviceConfiguration;
import org.openmuc.core.driver.DataReceivedListener;
import org.openmuc.core.driver.DeviceDriverManager;
import org.openmuc.core.driver.NoSuchDriverException;
import org.openmuc.core.measurement.Measurement;
import org.openmuc.core.measurement.State;

public class Reader implements Runnable, DataReceivedListener {
	private static Logger logger = Logger.getLogger(Reader.class);

	private DeviceDriverManager driverManager;
	
	private List<DeviceSamplingInfo> deviceSampInfo;

	private boolean running = false;
	
	private boolean finish = false;
	
	private Configuration configuration;
	
	public Reader(DeviceDriverManager driverManager, Configuration configuration, List<DeviceConfiguration> devices) {
		super();
		this.driverManager = driverManager;
		this.configuration = configuration;
		
		deviceSampInfo = new ArrayList<DeviceSamplingInfo>(devices.size());
		
		for (DeviceConfiguration device : devices) {
			deviceSampInfo.add(new DeviceSamplingInfo(device));
		}
	}
	
	public void run() {
		running = true;

		while (!finish) {
			
			for (DeviceSamplingInfo deviceInfo : deviceSampInfo) {
				
				DeviceConfiguration device = deviceInfo.getDeviceConfig();
				
				long sampleTime = System.currentTimeMillis();
				
				try {
					if ((deviceInfo.getLastRequestTime() == null)
						|| ((sampleTime >= (deviceInfo.getLastRequestTime() + device.getSamplingPeriod()))
						&& (deviceInfo.getLastRequestHandle().finished))
						)
					{											
						if (device.getUrl() != null) 
						{
							ReaderHandle handle = new ReaderHandle(device, configuration.getStoragesForMeter(device));
							
							try {
								deviceInfo.setLastRequestTime(sampleTime);
								deviceInfo.setLastRequestHandle(handle);
								driverManager.readDeviceAsync(device.getUrl(), device.getChannelIdentifierList(), this, handle);
								

							} catch (NoSuchDriverException e) {
								logger.debug(e.getMessage());
								logger.info("Please load driver for: " + device.getUrl().toString());
								finish = true;
							}
						}
					}
				} catch (Exception e) {
					logger.debug("Unhandled thread exception reading meter " +  
							device.getName() + ":" + e.getMessage());
					finish = true;
					e.printStackTrace();
				}
				
				
				if (finish) break;
				
			}
			
			try {Thread.sleep(10);} catch (InterruptedException e) {}
		}

		logger.info("readout process stopped!");
		
		running = false;
	}

	public void finish() {
		finish = true;
	}

	public boolean isRunning() {
		return running;
	}

	public void setRunning(boolean running) {
		this.running = running;
	}



	public void dataReceived(Object handle, List<Measurement> data) {
		insertIntoDataStore(((ReaderHandle) handle).getStorages(), data);
		
		((ReaderHandle) handle).finished = true;			
	}

	private void insertIntoDataStore(List<DataStorage> dataStorages,
			List<Measurement> data) 
	{		
		for (Measurement msr : data) {
						
			for (DataStorage dataStorage : dataStorages) {
				if (dataStorage.addMeasurement(msr)) {
					break;
				}
			}
		}		
	}

	public void readoutError(Object handle, int errorCode) {		
		ReaderHandle readerHandle = (ReaderHandle) handle;
		
		readerHandle.finished = true;
				
		logger.debug("Error reading device: " + readerHandle.getDevice().getUrl().toString()  
				+ " code: " + errorCode);
		
		for (DataStorage storage : readerHandle.getStorages()) {
			storage.setState(State.invalid);
		}
	}

}
