package org.openmuc.core.datamanager.impl;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.openmuc.core.datamanager.Authentication;
import org.openmuc.core.datamanager.DataManager;
import org.openmuc.core.datamanager.DataManagerListener;
import org.openmuc.core.datamanager.DataStorage;
import org.openmuc.core.datamanager.DatabaseInterface;
import org.openmuc.core.datamanager.PlantInfo;
import org.openmuc.core.datamanager.config.ChannelConfiguration;
import org.openmuc.core.datamanager.config.ConfigurationService;
import org.openmuc.core.datamanager.config.DeviceConfiguration;
import org.openmuc.core.datamanager.config.PlantConfiguration;
import org.openmuc.core.driver.DeviceDriverManager;
import org.osgi.framework.BundleContext;
import org.osgi.util.tracker.ServiceTracker;

/*
 * Central storage and data access point.
 * 
 * - Database handling
 * - Data storage
 * - Security management
 */

public class DataManagerImpl implements DataManager {
	private static Logger logger = Logger.getLogger(DataManagerImpl.class);

	private Vector<DataStorage> storages;

	private ConfigurationService configuration;

	private DatabaseInterface databaseLayer;

	private DeviceDriverManager manager;

	private ServiceTracker listenerTracker;

	private PlantInfo plantInfo = null;

	protected DataManagerImpl(BundleContext bc, DeviceDriverManager manager, ConfigurationService configuration) {
		this.manager = manager;

		/* Create service tracker for DataManagerListener */
		listenerTracker = new ServiceTracker(bc, DataManagerListener.class.getName(), null);
		listenerTracker.open();

		this.configuration = configuration;

		if (configuration != null) {
			PlantConfiguration plantConfig = configuration.getPlantConfiguration();

			if (plantConfig != null) {
				this.plantInfo = new PlantInfo(plantConfig.getName());
			}
		}

		/* Connect to database */
		this.databaseLayer = DatabaseLayerFactory.getInstance();

		restart();
	}

	protected void restart() {
		logger.debug("Create new data store");

		this.storages = new Vector<DataStorage>();

		/* Create data storages */
		createDataStoragesFromConfiguration(this.configuration.getDevices());

		/* Create new database registers from configuration */
		for (DeviceConfiguration device : configuration.getDevices()) {
			for (ChannelConfiguration channel : device.getChannels()) {
				try {
					if (channel.getStorageType() != ChannelConfiguration.STORE_TYPE_NONE)
						databaseLayer.addChannel(channel);
				} catch (IOException e) {
					logger.error("Failed to create storage for channel " + channel.getLabel(), e);
				}
			}
		}

		logger.info("Configured data store from configuration file");
	}

	private void createDataStoragesFromConfiguration(List<DeviceConfiguration> devices) {
		if (devices != null) {
			for (DeviceConfiguration device : devices) {
				for (ChannelConfiguration channel : device.getChannels()) {
					storages.add(new DataStorageImpl(listenerTracker, channel, this.databaseLayer, device.getUrl(),
							manager));
				}
			}
		}
	} /* createDataStoragesFromConfiguration() */

	protected void addDataStorage(DataStorage storage) {
		this.storages.add(storage);

		/* Inform listeners */
		DataManagerListener listener;
		Object[] services = listenerTracker.getServices();

		if (services != null) {
			for (Object service : services) {
				listener = (DataManagerListener) service;
				listener.updated();
			}
		}
	} /* addDataStorage() */

	public DataStorage getDataStorage(String label, Authentication auth) {

		for (DataStorage storage : storages) {
			if (storage.getDataId().getLabel().equals(label)) {
				return storage;
			}
		}
		return null;
	}

	public List<String> getDeviceDirectory() {
		LinkedList<String> idList = new LinkedList<String>();

		for (DataStorage storage : storages) {
			idList.add(storage.getDataId().getLabel());
		}

		return idList;
	} /* getDeviceDirectory() */

	protected List<DataStorage> getDataStorages() {
		return storages;
	}

	@Override
	protected void finalize() throws Throwable {
		listenerTracker.close();
		super.finalize();
	}

	public PlantInfo getPlantInfo() {
		return plantInfo;
	}

}
