package org.openmuc.core.datamanager.impl;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.IOException;
import java.lang.ref.WeakReference;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.openmuc.core.datamanager.DataManagerListener;
import org.openmuc.core.datamanager.DataStorageListener;
import org.openmuc.core.datamanager.DatabaseInterface;
import org.openmuc.core.datamanager.config.ChannelConfiguration;
import org.openmuc.core.driver.DeviceDriverManager;
import org.openmuc.core.driver.DeviceURI;
import org.openmuc.core.measurement.Measurement;
import org.openmuc.core.measurement.ChannelIdentifier;
import org.openmuc.core.measurement.State;
import org.openmuc.core.measurement.Unit;
import org.openmuc.core.measurement.Value;
import org.osgi.util.tracker.ServiceTracker;

public class DataStorageImpl implements org.openmuc.core.datamanager.DataStorage {
	
	/* Use WeakReference so listeners don't need to unregister */
	Set<WeakReference<DataStorageListener>> listeners;
	
	private static Logger logger = Logger.getLogger(DataStorageImpl.class);
	
	private ChannelIdentifier dataId = null;
	private Vector<Value> data = null;
	private Value currentValue = null;
	private DeviceDriverManager driverManager;
	private ServiceTracker listenerTracker;
	
	private ChannelConfiguration channelConfig;
	private DatabaseInterface database = null;
	
	private DeviceURI deviceURL;
	
	private long storingPeriod;
	private long nextStoringTime;
	private int storingMethod;
	
	private Double lowerLimit = null;
	private Double upperLimit = null;
	
	private long lastUpdateTimestamp = 0;
	
	private Double lastStoredValue = null;
	
	private State state = State.valid;
			
	public DataStorageImpl(ServiceTracker listenerTracker, ChannelConfiguration channelConfig, 
			DatabaseInterface database, DeviceURI url, DeviceDriverManager driverManager)
	{
		this.channelConfig = channelConfig;
		this.database = database;
		data = new Vector<Value>();
		this.dataId = channelConfig.getDataID();
		this.storingMethod = channelConfig.getStorageType();
		this.deviceURL = url;
		this.driverManager = driverManager;
		this.listenerTracker = listenerTracker;
		this.lowerLimit = channelConfig.getLowerLimit();
		this.upperLimit = channelConfig.getUpperLimit();
		
		this.listeners = new LinkedHashSet<WeakReference<DataStorageListener>>();
		
		if ((this.storingMethod == ChannelConfiguration.STORE_TYPE_INST) || 
				(this.storingMethod == ChannelConfiguration.STORE_TYPE_MEAN)) 
		{
			calculateStoringValues();
		}
		
		currentValue = null;		
	}
	
	private void calculateStoringValues() {		
		long startTime = channelConfig.getStoringStartTime().getTime();
		long period = channelConfig.getStoringPeriod() * 1000;
		long curTime = System.currentTimeMillis();		
		long nbOfPeriodsSinceRef = (curTime - startTime) / period;
		
		long nextTime = startTime + ((nbOfPeriodsSinceRef + 1) * period);
		
		this.storingPeriod = period;
		this.nextStoringTime = nextTime;
	}
	
	private long calculateNextStoringTime() {
		long startTime = channelConfig.getStoringStartTime().getTime();
		long curTime = System.currentTimeMillis();		
		long nbOfPeriodsSinceRef = (curTime - startTime) / this.storingPeriod;
		
		return startTime + ((nbOfPeriodsSinceRef + 1) * this.storingPeriod);
	}
	
	public Value getCurrentValue() {
		return currentValue;
	}
	
	public void setDatabase(DatabaseInterface dbl) {
		this.database = dbl;
	}

	public boolean addMeasurement(Measurement msr) {
		if (this.dataId == msr.getDataID()) {
			this.lastUpdateTimestamp = System.currentTimeMillis();
			
			double value = (msr.getValue().getValue() * channelConfig.getScaler()) + 
				channelConfig.getOffset();
			
			
			State newState = State.valid;
			
			/* Check limits */
			if (lowerLimit != null) {
				if (value < lowerLimit) newState = State.lower_limit_exceeded;
			}
			if (upperLimit != null) {
				if (value > upperLimit) newState = State.upper_limit_exceeded;
			}
			
			if (newState != this.state) {
				this.state = newState;
				informStateChangedListeners();
			}
						
			this.currentValue = new Value(value, msr.getValue().getTimestamp());
					
			storeData();
			informDataStorageListeners();
			
			return true;
		}
		else return false;
	}

	private void storeData() {
		long currentTime = System.currentTimeMillis();

		switch (this.storingMethod) {
		
		case ChannelConfiguration.STORE_TYPE_NONE:
			break;
			
		case ChannelConfiguration.STORE_TYPE_INST:
			if (currentTime >= this.nextStoringTime) {
				long nextTime = calculateNextStoringTime();
				
				storeInDatabase(this.currentValue.getValue(), nextTime - this.storingPeriod);
				
				this.nextStoringTime = nextTime;
			} 
			break;
			
		case ChannelConfiguration.STORE_TYPE_MEAN:
			
			data.add(currentValue);
			
			if (currentTime >= this.nextStoringTime) {
				long nextTime = calculateNextStoringTime();
				double meanValue = calculateMeanValue(nextTime - (2 * this.storingPeriod));
				this.data = new Vector<Value>();
				
				storeInDatabase(meanValue, nextTime - this.storingPeriod);
			
				this.nextStoringTime = nextTime;
			} 

			break;
			
		case ChannelConfiguration.STORE_TYPE_VALUE_CHANGED:
			if (shouldValueBeStored()) {		
				storeInDatabase(this.currentValue.getValue(), this.currentValue.getTimestamp());
				lastStoredValue = this.currentValue.getValue();
			}		
			break;
		}
	}
	
	private boolean shouldValueBeStored() {	
		if (lastStoredValue == null) {return true;}
		
		if (channelConfig.getAbsoluteChangeLevel() != null) {
			if (Math.abs(this.currentValue.getValue() - lastStoredValue) >= Math.abs(channelConfig.getAbsoluteChangeLevel()))
					return true;
		}

		return false;
	}
	
	private void storeInDatabase(double value, long storingTime) {
		try {
			this.database.insertValue(this.dataId.getLabel(),new Value(value, storingTime));
		} catch (IOException e) {
			logger.error(e.getMessage());
		}	
		
		informStoringListeners(value);
	}

	private void informStoringListeners(double value) {
		DataManagerListener listener;
		Object[] services = listenerTracker.getServices();

		if (services != null) {
			for (Object service : services) {
				listener = (DataManagerListener) service;
				listener.storingDataEvent(dataId.getLabel(),
						this.nextStoringTime, value);
			}
		}
	}
	
	private void informStateChangedListeners() {
		DataManagerListener listener;
		Object[] services = listenerTracker.getServices();

		if (services != null) {
			for (Object service : services) {
				listener = (DataManagerListener) service;
				listener.stateChanged(dataId, state);
			}
		}
	}
	
	private void informDataStorageListeners() {
		for (WeakReference<DataStorageListener> reference : listeners) {
			if (reference.get() != null) {
				DataStorageListener listener = reference.get();
				
				listener.updated(dataId.getLabel(), lastUpdateTimestamp, currentValue, state);
			}
			else listeners.remove(reference);
		}
	}
	
	private double calculateMeanValue(long periodStart /* averaging start time*/ ) {
		int valueCount = 0;
		double meanValue = 0;
		double lastMeanValue = 0;
		
		for (Value value : data) {
			if (value.getTimestamp() >= periodStart) {
				valueCount++;
				meanValue = lastMeanValue - ((lastMeanValue - value.getValue()) / valueCount);
				lastMeanValue = meanValue;
			}
		}
		
		return meanValue;
	} /* calculateMeanValue() */
	
	public double getMeanValue() {
		return calculateMeanValue(calculateNextStoringTime() - (2* this.storingPeriod));
	}

	public ChannelIdentifier getDataId() {
		return dataId;
	}

	public Value getValue(long timestamp) {
		Value value;
		
		try {
			value = this.database.getValue(this.dataId.getLabel(), timestamp);
		} catch (IOException e) {
			
			value = null;
		}
		
		return value;
	}

	public List<Value> getValues(long startTime) {
		try {
			if ((this.database == null) || (this.dataId == null)) {
				System.out.println("nullpointer") ;
				return null;	
			}
			return this.database.getValues(this.dataId.getLabel(), startTime);
		} catch (IOException e) {
			logger.error(e.getMessage());
			return new LinkedList<Value>();
		}
	}

	public List<Value> getValues(long startTime, long endTime) {
		try {
			return this.database.getValues(this.dataId.getLabel(), startTime, endTime);
		} catch (IOException e) {
			logger.error(e.getMessage());
			return new LinkedList<Value>();
		}
	}

	public long getLastUpdateTime() {
		return this.lastUpdateTimestamp;
	}

	public void setValue(double value) {
		try {		
			double writeVal = (value  - channelConfig.getOffset()) / channelConfig.getScaler();
			
			if (this.getDataId().getObjectLocator() != null
					&& !this.getDataId().getObjectLocator().equals("")) 
			{
				this.state = State.valid;
				driverManager.writeChannel(deviceURL, this.dataId, writeVal);
				this.currentValue = new Value(value, System.currentTimeMillis());
			}
			else {
				this.addMeasurement(new Measurement(dataId, new Value(value, System.currentTimeMillis()), Unit.UNKNOWN));
			}
		} catch (Exception e) {
			logger.error("setValue failed for " + this.dataId.getLabel());
		}
	}

	public State getState() {
		return state;
	}

	public void addListener(DataStorageListener listener) {
		listeners.add(new WeakReference<DataStorageListener>(listener));
	}

	public String getLabel() {
		return dataId.getLabel();
	}

	public void setState(State state) {
		this.state = state;
	}

	public Double getLowerLimit() {
		return lowerLimit;
	}

	public Double getUpperLimit() {
		return upperLimit;
	}

	public ChannelConfiguration getConfiguration() {
		return channelConfig;
	}
		
}
