/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.openmuc.core.datamanager.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.openmuc.core.datamanager.config.DeviceConfiguration;
import org.openmuc.core.datamanager.config.ChannelConfiguration;
import org.openmuc.core.driver.DeviceURI;
import org.openmuc.core.measurement.ChannelIdentifier;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * 
 * @author mzillgit
 *
 */
public class DeviceConfigurationImpl implements DeviceConfiguration {
	private static Logger logger = Logger.getLogger(DeviceConfigurationImpl.class);
	
	private String name;
	private String meterID = null;
	private String vendorID = null;
	private String deviceID = null;
	private String version = null;
	private String description = null;
	private Long samplingPeriod = null;

	private DeviceURI url = null;
	private List<ChannelConfiguration> channels;
	private List<ChannelIdentifier> dataIDList = null; /* only a buffer to increase performance */
	
	private HashMap<String,String> properties = new HashMap<String,String>();
	
	private Configuration configuration;
	
	protected DeviceConfigurationImpl(Configuration configuration, String name) {
		this.configuration = configuration;
		this.name = name;
		channels = new ArrayList<ChannelConfiguration>();
	}
	
	protected DeviceConfigurationImpl (Configuration configuration, Node nodeMeter) {
		this.configuration = configuration;
		List<ChannelConfiguration> channels = new ArrayList<ChannelConfiguration>();
		
		NodeList childNodes = nodeMeter.getChildNodes();
	
		NamedNodeMap attributes = nodeMeter.getAttributes();
		
		Node nameAttribute = attributes.getNamedItem("name");
		
		if (nameAttribute != null) {
			this.name = nameAttribute.getTextContent();
		}
		
		for (int i = 0; i<childNodes.getLength(); i++) {
			Node child = childNodes.item(i);
			if(child.getNodeName().equals("#text") || child.getNodeName().equals("#comment"))
				continue;
			else if(child.getNodeName().equals("meterID"))
				this.meterID = child.getTextContent();
			else if(child.getNodeName().equals("vendorID"))
				this.vendorID = child.getTextContent();
			else if(child.getNodeName().equals("deviceID"))
				this.deviceID = child.getTextContent();
			else if(child.getNodeName().equals("version"))
				this.version = child.getTextContent();
			else if(child.getNodeName().equals("comString"))
				this.url = new DeviceURI(child.getTextContent());
			else if(child.getNodeName().equals("samplingPeriod"))
				this.samplingPeriod = XMLUtils.timeToMilliSec(child.getTextContent());
			else if(child.getNodeName().equals("description"))
				this.description = child.getTextContent();
			else if (child.getNodeName().equals("property"))
				addNewProperty(child);
			else if (child.getNodeName().equals("register") || 
					child.getNodeName().equals("channel")) 
			{
				ChannelConfiguration channel = new ChannelConfigurationImpl(child);
				logger.info("NEW CHANNEL: " + channel.getLabel());
				channels.add(channel);
			}
			else
				logger.info(child.getNodeName() +  " is no valid entry");
		}
		
		this.channels = channels;
	}
	
	private void addNewProperty(Node child) {
		String propertyName = XMLUtils.getAttributeValue(child, "name");
		
		if (propertyName == null) {
			logger.info("Not a valid property!");
			return;
		}
		
		String value = child.getTextContent();
		
		if (value == null) {
			logger.info("property value is missing!");
			return;
		}
		
		properties.put(propertyName, value);
	}

	public ChannelConfiguration createChannel(String label) {
		dataIDList = null;
		ChannelConfigurationImpl configuration = new ChannelConfigurationImpl(label);
		channels.add(configuration);
		return configuration;
	}
	
	public void deleteChannel(String label) {
		ChannelConfiguration register = getChannelByLabel(label);
		channels.remove(register);
		XMLUtils.deleteElementWithSubElementValue(configuration.getXmlDocument(), 
				"channel", "label", label);
		XMLUtils.deleteElementWithSubElementValue(configuration.getXmlDocument(), 
				"register", "label", label);
	}
	
	public String getMeterID() {
		return meterID;
	}
	
	public String getVendorID() {
		return vendorID;
	}
	
	public String getDeviceID() {
		return deviceID;
	}

	public String getVersion() {
		return version;
	}

	public DeviceURI getUrl() {
		return url;
	}

	public List<ChannelConfiguration> getChannels() {
		return Collections.unmodifiableList(channels);
	}
	
	public ChannelConfiguration getChannelByLabel(String label) {
		for (ChannelConfiguration channel : channels) {
			if (channel.getLabel().equals(label))
				return channel;
		}
		
		return null;
	}
	
	public List<ChannelIdentifier> getChannelIdentifierList() {
		
		if (this.dataIDList == null) {
			this.dataIDList = new ArrayList<ChannelIdentifier>();
			
			for (ChannelConfiguration channel : channels) {
				this.dataIDList.add(channel.getDataID());
			}
		}
		
		return this.dataIDList;
	} /* getDataIDList() */

	public String getDescription() {
		return description;
	}

	public void setDescription(String description) {
		this.description = description;
	}

	public void setVendorID(String vendorID) {
		this.vendorID = vendorID;
	}

	public void setDeviceID(String deviceID) {
		this.deviceID = deviceID;
	}

	public void setVersion(String version) {
		this.version = version;
	}

	protected void persist(Element parent, Document document) {
		Element device = null;

		device = (Element) XMLUtils.hasTagAsChildWithAttributeValue(parent, 
				"meter", "name", this.name);
		
		device = (Element) XMLUtils.hasTagAsChildWithAttributeValue(parent, 
				"device", "name", this.name);
		
		if (device == null) {
			device = document.createElement("device");
			device.setAttribute("name", this.name);
			parent.appendChild(device);
		}
		
		XMLUtils.updateTagWithValue(device, document, "description", this.description);
		XMLUtils.updateTagWithValue(device, document, "meterID", this.meterID);
		XMLUtils.updateTagWithValue(device, document, "deviceID", this.deviceID);

		if (this.url != null) {
			XMLUtils.updateTagWithValue(device, document, "comString", this.url.toString());
		}
		
		XMLUtils.updateTagWithValue(device, document, "vendorID", this.vendorID);
		XMLUtils.updateTagWithValue(device, document, "version", this.version);
	
		if (this.samplingPeriod != null)
			XMLUtils.updateTagWithValue(device, document, "samplingPeriod", this.samplingPeriod.toString());
		
		// store properties
		Iterator<String> propertyList =  properties.keySet().iterator();
		
		while (propertyList.hasNext()) {
			String name = propertyList.next();
			String value = properties.get(name);
			
			Element propertyTag = (Element) XMLUtils.hasTagAsChildWithAttributeValue(device, 
					"property", "name", name);
			
			if (propertyTag == null) {
				propertyTag = document.createElement("property");
				propertyTag.setAttribute("name", name);
			}
			
			propertyTag.setTextContent(value);
			
			device.appendChild(propertyTag);
		}
		
		for (ChannelConfiguration channel : channels) {
			((ChannelConfigurationImpl) channel).persist(device, document);
		}
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public void setMeterUrl(String meterUrl) {
		this.url = new DeviceURI(meterUrl);
	}

	public String getProperty(String name) {
		return properties.get(name);
	}

	public void setProperty(String name, String value) {
		properties.put(name, value);
	}

	public void deleteProperty(String name) {
		properties.remove(name);
	}

	public Map<String, String> getPropertyList() {
		return properties;
	}

	public long getSamplingPeriod() {
		if (this.samplingPeriod == null) 
			return 1000; /* default sampling rate = 1000 ms */
		else 
			return this.samplingPeriod;
	}

	public void setSamplingPeriod(long period) {
		this.samplingPeriod = new Long(period);
	}
	
}
