package org.openmuc.core.datamanager.test;

import java.io.File;
import java.util.List;

import org.junit.AfterClass;
import org.junit.Test;
import org.openmuc.core.datamanager.config.ConfigurationService;
import org.openmuc.core.datamanager.config.DeviceConfiguration;
import org.openmuc.core.datamanager.config.ChannelConfiguration;
import org.openmuc.core.datamanager.config.PlantConfiguration;
import org.openmuc.core.datamanager.impl.Configuration;
import org.openmuc.core.measurement.ChannelIdentifier;

import static org.junit.Assert.*;

public class TestConfiguration {
	
	@Test
	public void testLoadingConfigurationFile() {
		ConfigurationService configuration = new Configuration("src/test/resources/measurements.xml");
		
		assertNotNull(configuration);
		
		List<DeviceConfiguration> meters = configuration.getDevices();
		
		assertNotNull(meters);
		
		assertEquals(2, meters.size());
		
		DeviceConfiguration meterConfig = meters.get(0);
		
		assertNotNull(meterConfig);
		
		List<ChannelConfiguration> registers = meterConfig.getChannels();
		
		assertNotNull(registers);
		
		assertEquals(2, registers.size());
		
		ChannelConfiguration register = registers.get(0);
		
		assertNotNull(register);
		
		assertEquals("elec.energy", register.getDataID().getLabel());
		assertEquals(60, register.getStoringPeriod());
		assertEquals(ChannelConfiguration.STORE_TYPE_INST, register.getStorageType());
		assertEquals("mbus:04/03", register.getObjectLocator());
		
		ChannelIdentifier dataId = register.getDataID();
		assertNotNull(dataId);
		assertEquals(register.getObjectLocator(), dataId.getObjectLocator());
		assertEquals(register.getLabel(), dataId.getLabel());
		
		/* Check dummy meter - without comString and objectlocators */
		meterConfig = meters.get(1);
		assertNotNull(meterConfig);
		
		List<ChannelIdentifier> dataIdList = meterConfig.getChannelIdentifierList();
		assertEquals(3, dataIdList.size());
		assertNotNull(dataIdList.get(0));
		assertNotNull(dataIdList.get(0).getLabel());
	}
	
	@Test
	public void testLoadingConfigurationFile2() {
		ConfigurationService configuration = new Configuration("src/test/resources/meterconfig.xml");
		
		assertNotNull(configuration);
		
		List<DeviceConfiguration> devices = configuration.getDevices();
		
		assertNotNull(devices);
		
		assertEquals(1, devices.size());
		
		DeviceConfiguration deviceConfiguration = devices.get(0);
		
		assertNotNull(deviceConfiguration);
		
		List<ChannelConfiguration> channels = deviceConfiguration.getChannels();
		
		assertNotNull(channels);
		
		assertEquals(2, channels.size());
		
		ChannelConfiguration channel = channels.get(0);
		
		assertNotNull(channel);
		
		assertEquals("elec.energy", channel.getDataID().getLabel());
		assertEquals(60, channel.getStoringPeriod());
		assertEquals(ChannelConfiguration.STORE_TYPE_INST, channel.getStorageType());
		assertEquals("mbus:04/03", channel.getObjectLocator());
		
		ChannelIdentifier dataId = channel.getDataID();
		assertNotNull(dataId);
		assertEquals(channel.getObjectLocator(), dataId.getObjectLocator());
		assertEquals(channel.getLabel(), dataId.getLabel());
		
		/* Check dummy meter - without comString and objectlocators */
		deviceConfiguration = devices.get(0);
		assertNotNull(deviceConfiguration);
		
		List<ChannelIdentifier> dataIdList = deviceConfiguration.getChannelIdentifierList();
		assertEquals(2, dataIdList.size());
		assertNotNull(dataIdList.get(0));
		assertNotNull(dataIdList.get(0).getLabel());
		
		PlantConfiguration plantConfig = configuration.getPlantConfiguration();
		
		assertNotNull(plantConfig);
		assertEquals("power.plant", plantConfig.getName());
	}
	
	@Test
	public void testCreationOfConfigurationFile() {

		deleteFileIfExists("testconfig.xml");
		ConfigurationService configuration = new Configuration("testconfig.xml");
		
		assertNotNull(configuration);
		
		DeviceConfiguration meterConfig = configuration.createDevice("electricity.meter");
		meterConfig.setDescription("A test meter");
		meterConfig.setVendorID("TEST");
		meterConfig.setVersion("1");
		
		ChannelConfiguration registerConfig = meterConfig.createChannel("voltage");
			
		assertNotNull(registerConfig);
		
		registerConfig = meterConfig.createChannel("current");
		
		registerConfig.setScaler(2.);
		
		registerConfig.setOffset(3.);
		
		configuration.persist();
		
		configuration = null;
		System.gc();
		
		File configFile = new File("testconfig.xml");

		assertEquals(true, configFile.exists());
		
		configuration = new Configuration("testconfig.xml");
		
		meterConfig = configuration.getDeviceByName("electricity.meter");
		
		assertNotNull(meterConfig);
		
		meterConfig.createChannel("power");
		
		configuration.persist();
		
		configuration = null;
		System.gc();
		
		configuration = new Configuration("testconfig.xml");
		
		/* Check if configuration is correct */
		List<DeviceConfiguration> meters = configuration.getDevices();
		assertNotNull(meters);
		assertEquals(1, meters.size());
		testMeterAndLabels(configuration);
		
	}
	
	@Test
	public void testManipulationOfConfigurationFile() {
		/* Ensure that meter exits in configuration */
		ConfigurationService configuration = new Configuration("testconfig.xml");
		DeviceConfiguration meterConfig = configuration.getDeviceByName("electricity.meter");
		assertNotNull(meterConfig);
		
		List<ChannelConfiguration> registers = meterConfig.getChannels();
		ChannelConfiguration register = meterConfig.getChannelByLabel("power");
		assertNotNull(register);
		
		/* Deleting a register from the registers list should not be possible */
		try {
			registers.remove(register);
			assertTrue(false);
		} catch (UnsupportedOperationException e) {
		}
		
		/* Delete register from meter */
		meterConfig.deleteChannel("power");
		register = meterConfig.getChannelByLabel("power");
		assertNull(register);
		
		configuration.persist();
		
		/* Reload meter file to test if deletion of register was successfull */
		configuration = new Configuration("testconfig.xml");
		meterConfig = configuration.getDeviceByName("electricity.meter");
		assertNotNull(meterConfig);
		register = meterConfig.getChannelByLabel("power");
		assertNull(register);
		
		/* Delete meter from configuration */
		configuration.deleteDevice("electricity.meter");
		configuration.persist();
		
		/* Reload configuration file */
		configuration = new Configuration("testconfig.xml");
		meterConfig = configuration.getDeviceByName("electricity.meter");
		assertNull(meterConfig);
	}

	private void testMeterAndLabels(ConfigurationService configuration) {
		DeviceConfiguration device;
		device = configuration.getDeviceByName("electricity.meter");
		
		assertNotNull(device);	
		assertEquals("A test meter", device.getDescription());
		assertEquals("TEST", device.getVendorID());
		assertEquals("1", device.getVersion());
		assertNotNull(device.getChannels());
		assertEquals(3, device.getChannels().size());
		
		assertNotNull(device.getChannelByLabel("current"));
		assertNotNull(device.getChannelByLabel("voltage"));
		assertNotNull(device.getChannelByLabel("power"));
		
		ChannelConfiguration channel = device.getChannelByLabel("current");
		
		assertEquals(2.0, channel.getScaler(), 0.01);
		assertEquals(3.0, channel.getOffset(), 0.01);
	}
	
	@Test
	public void testPropertyHandling() {
		deleteFileIfExists("testconfig.xml");
		ConfigurationService configuration = new Configuration("testconfig.xml");
		
		DeviceConfiguration device = configuration.createDevice("cxplant");
		
		device.setDescription("general plant properties");
		
		device.setProperty("name", "test tracker");
		device.setProperty("wind_limit", "15");
		
		configuration.persist();
		
		configuration = new Configuration("testconfig.xml");
		
		device = configuration.getDeviceByName("cxplant");
		
		assertNotNull(device);
		assertEquals("test tracker", device.getProperty("name"));
		assertEquals("15", device.getProperty("wind_limit"));
		assertNull(device.getProperty("max"));
		device.deleteProperty("wind_limit");
		assertNull(device.getProperty("wind_limit"));
		
		
	}
	
	private static void deleteFileIfExists(String filename) {
		File testFile = new File(filename);
		if (testFile.exists()) {
			testFile.delete();
		}
	}
	
	@AfterClass
	public static void cleanUp() {
		deleteFileIfExists("testconfig.xml");
	}
	
}
