package org.openmuc.metering.driver.manager;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.IOException;
import java.util.Hashtable;
import java.util.List;

import org.apache.log4j.Logger;
import org.openmuc.core.driver.BusURI;
import org.openmuc.core.driver.ChannelInformation;
import org.openmuc.core.driver.DataReceivedListener;
import org.openmuc.core.driver.DeviceDriverInterface;
import org.openmuc.core.driver.DeviceDriverManager;
import org.openmuc.core.driver.DeviceURI;
import org.openmuc.core.driver.MethodNotImplementedException;
import org.openmuc.core.driver.NoSuchDriverException;
import org.openmuc.core.measurement.ChannelIdentifier;
import org.openmuc.core.measurement.Measurement;

import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.osgi.util.tracker.ServiceTracker;

/**
 * 
 * 
 * 
 * @author mzillgit
 *
 */
public class DriverManagerImpl implements DeviceDriverManager {
	private static Logger logger = Logger.getLogger(DriverManagerImpl.class);
	
	private Hashtable<String, DeviceDriverInterface> drivers;
	
	class MeterDriverTracker extends ServiceTracker {
		
		public MeterDriverTracker(BundleContext context) {
			super(context, DeviceDriverInterface.class.getName(), null);
		}

		@Override
		public Object addingService(ServiceReference reference) {
			DeviceDriverInterface driver = (DeviceDriverInterface) context.getService(reference);
			
			registerDriver(driver, driver.getDriverID());
			
			return driver;
		}

		@Override
		public void removedService(ServiceReference reference, Object service) {
			unregisterDriver((DeviceDriverInterface) service);
			
			context.ungetService(reference);
		}
		
		
	} /* class MeterDriverTracker */
	
	ServiceTracker getServiceTracker(BundleContext context) {
		return new MeterDriverTracker(context);
	}
	
	public DriverManagerImpl() {
		drivers = new Hashtable<String, DeviceDriverInterface>();
	}
	
	/* (non-Javadoc)
	 * @see org.openmuc.metering.driver.manager.MeterDriverManager#registerDriver(org.openmuc.metering.driver.manager.MeterDriverInterface, java.lang.String)
	 */
	public void registerDriver(DeviceDriverInterface driver, String driverID) {
		drivers.put(driverID, driver);
		logger.info("registerDriver(): registered new driver " + driverID);
	}
	
	/* (non-Javadoc)
	 * @see org.openmuc.metering.driver.manager.MeterDriverManager#unregisterDriver(org.openmuc.metering.driver.manager.MeterDriverInterface)
	 */
	public void unregisterDriver(DeviceDriverInterface driver) {
		String key;
		
		key = driver.getDriverID();
		
		if (drivers.get(key) == driver) {
			drivers.remove(key);
		}
		else {
			logger.info("unregisterDriver(): No such driver registered!");
		}
	} /* unregisterDriver() */
	
	/* (non-Javadoc)
	 * @see org.openmuc.metering.driver.manager.MeterDriverManager#readMeterAsync(org.openmuc.metering.driver.manager.MeterURL, java.util.Vector, org.openmuc.metering.driver.manager.MeterDataReceivedListener, java.lang.Object)
	 */
	public void readDeviceAsync(DeviceURI url, List<ChannelIdentifier> registers, 
			DataReceivedListener callback, Object handle) 
		throws NoSuchDriverException 
	{
		DeviceDriverInterface driver = null;
		String driverID;
		
		driverID = url.getDriverID();
		
		driver = drivers.get(driverID);
		
		if (driver != null) {
			try {
				driver.readDeviceAsync(url, registers, callback, handle);
			} catch (MethodNotImplementedException e) {
				// TODO Auto-generated catch block
				throw new NoSuchDriverException(driverID + " does not support this operation!");
			}
		}
		else throw new NoSuchDriverException(driverID);
	} /* readMeterAsync() */
	
	/* (non-Javadoc)
	 * @see org.openmuc.metering.driver.manager.MeterDriverManager#writeRegister(org.openmuc.metering.driver.manager.MeterURL, org.openmuc.metering.MeterDataIdentifier, double)
	 */
	public void writeChannel(DeviceURI url, ChannelIdentifier register, double value) 
		throws NoSuchDriverException, MethodNotImplementedException, IOException
	{
		DeviceDriverInterface driver = null;
		String driverID;
		
		driverID = url.getDriverID();
		
		driver = drivers.get(driverID);
		
		if (driver != null) {
			driver.writeChannel(url, register, value);
		}
	} /* writeChannel() */

	public List<DeviceURI> getConnectedDevices(BusURI url)
			throws MethodNotImplementedException, NoSuchDriverException,
			IOException {
		List<DeviceURI> toReturn = null;
		
		DeviceDriverInterface driver = null;
		String driverID;
		
		driverID = url.getDriverID();
		
		driver = drivers.get(driverID);
		
		if (driver != null) {
			toReturn = driver.getDeviceList(url);
		}
		
		return toReturn;
	}
	
	public List<ChannelInformation> getDeviceChannels(DeviceURI url)
			throws MethodNotImplementedException, NoSuchDriverException,
			IOException {
		List<ChannelInformation> toReturn = null;
		
		DeviceDriverInterface driver = null;
		String driverID;
		
		driverID = url.getDriverID();
		
		driver = drivers.get(driverID);
		
		if (driver != null) {
			toReturn = driver.getDeviceChannels(url);
		}
		
		return toReturn;
	}
	
	public List<Measurement> readDeviceSync(DeviceURI url, List<ChannelIdentifier> channels)
	throws MethodNotImplementedException, IOException {
		List<Measurement> toReturn = null;
		
		DeviceDriverInterface driver = null;
		String driverID;
		
		driverID = url.getDriverID();
		
		driver = drivers.get(driverID);
		
		if (driver != null) {
			toReturn = driver.readDeviceSync(url, channels);
		}
		
		return toReturn;
	}
	
} /* class MeterDriverManagerImpl */
