package org.openmuc.mux.dbprovider.slotsdb;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Simon Fey
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.Collections;
import java.util.List;
import java.util.Vector;

import org.openmuc.core.measurement.State;
import org.openmuc.core.measurement.Value;

public class FileObject {
	private long startTimeStamp; // byte 0-7 in file (cached)
	private long storagePeriod; // byte 8-15 in file (cached)
	private File dataFile;
	private DataOutputStream dos;
	private BufferedOutputStream bos;
	private FileOutputStream fos;
	private DataInputStream dis;
	private FileInputStream fis;
	private boolean canWrite;
	private boolean canRead;
	/*
	 * File length will be cached to avoid system calls an improve I/O
	 * Performance
	 */
	private long length = 0;

	/*
	 * This boolean sets the extended mode (if stepfrequency = -1) Which means,
	 * every measurement value has a timestamp!
	 */
	private boolean isExtendedFormat;

	public FileObject(String filename) throws IOException {
		canWrite = false;
		canRead = false;
		dataFile = new File(filename);
		this.length = dataFile.length();
		if (dataFile.exists() && this.length >= 16) {
			/*
			 * File already exists -> get file Header (startTime and
			 * step-frequency) TODO: compare to starttime and frequency in
			 * constructor! new file needed? update to file-array!
			 */
			try {
				fis = new FileInputStream(dataFile);
				try {
					dis = new DataInputStream(fis);
					try {
						this.startTimeStamp = dis.readLong();
						this.storagePeriod = dis.readLong();
					} finally {
						if (dis != null) {
							dis.close();
							dis = null;
						}
					}
				} finally {
					if (dis != null) {
						dis.close();
						dis = null;
					}
				}
			} finally {
				if (fis != null) {
					fis.close();
					fis = null;
				}
			}
		}
	}

	public FileObject(File file) throws IOException {
		canWrite = false;
		canRead = false;
		dataFile = file;
		this.length = dataFile.length();
		if (dataFile.exists() && this.length >= 16) {
			/*
			 * File already exists -> get file Header (startTime and
			 * step-frequency)
			 */
			fis = new FileInputStream(dataFile);
			try {
				dis = new DataInputStream(fis);
				try {
					this.startTimeStamp = dis.readLong();
					this.storagePeriod = dis.readLong();
				} finally {
					if (dis != null) {
						dis.close();
						dis = null;
					}
				}
			} finally {
				if (fis != null) {
					fis.close();
					fis = null;
				}
			}

		}
	}

	private void enableOutput() throws IOException {
		/*
		 * Close Input Streams, for enabling output.
		 */
		if (dis != null) {
			dis.close();
			dis = null;
		}
		if (fis != null) {
			fis.close();
			fis = null;
		}

		/*
		 * enabling output
		 */
		if (fos == null || dos == null || bos == null) {
			fos = new FileOutputStream(dataFile, true);
			bos = new BufferedOutputStream(fos);
			dos = new DataOutputStream(bos);
		}
		canRead = false;
		canWrite = true;
	}

	private void enableInput() throws IOException {
		/*
		 * Close Output Streams for enabling input.
		 */
		if (dos != null) {
			dos.flush();
			dos.close();
			dos = null;
		}
		if (bos != null) {
			bos.close();
			bos = null;
		}
		if (fos != null) {
			fos.close();
			fos = null;
		}

		/*
		 * enabling input
		 */
		if (fis == null || dis == null) {
			fis = new FileInputStream(dataFile);
			dis = new DataInputStream(fis);
		}
		canWrite = false;
		canRead = true;
	}

	/**
	 * Return the Timestamp of the first stored Value in this File.
	 */
	public long getStartTimeStamp() {
		return this.startTimeStamp;
	}

	/**
	 * @return step frequency in seconds
	 */
	public long getStoringPeriod() {
		return this.storagePeriod;
	}

	/**
	 * creates the file, if it doesn't exist.
	 * 
	 * @param startTimeStamp
	 *            for file header
	 */
	public void createFileAndHeader(long startTimeStamp, long stepIntervall) throws IOException {
		if (!dataFile.exists() || this.length < 16) {
			dataFile.getParentFile().mkdirs();
			if (dataFile.exists() && this.length < 16)
				dataFile.delete(); // file corrupted (header shorter that 16
			// bytes)
			dataFile.createNewFile();
			this.startTimeStamp = startTimeStamp;
			this.storagePeriod = stepIntervall;

			/*
			 * Do not close Output streams, because after writing the header ->
			 * data will follow!
			 */
			fos = new FileOutputStream(dataFile);
			bos = new BufferedOutputStream(fos);
			dos = new DataOutputStream(bos);
			dos.writeLong(startTimeStamp);
			dos.writeLong(stepIntervall);
			dos.flush();
			this.length += 16; /* wrote 2*8 Bytes */
			canWrite = true;
			canRead = false;
		}
	}

	public void append(double value, long timestamp, byte flag) throws IOException {
		long writePosition = getBytePosition(timestamp);
		if (writePosition == this.length) {
			/*
			 * value for this timeslot has not been saved yet "AND" some value
			 * has been stored in last timeslot
			 */
			if (!canWrite)
				enableOutput();
			// System.out.println("Appending: Timestamp = " + timestamp +
			// " Value = " + value + " Flag = " + flag);
			dos.writeDouble(value);
			dos.writeByte(flag);
			this.length += 9;
		}
		else {
			if (this.length > writePosition) {
				// System.out.println("Any other Value has already been stored for this timeslot: "
				// + timestamp + "!");
				/*
				 * value has already been stored for this timeslot -> handle?
				 * AVERAGE, MIN, MAX, LAST speichern?!
				 */
			}
			else {
				/*
				 * there are missing some values missing -> fill up with NaN!
				 */
				if (!canWrite)
					enableOutput();
				long rowsToFillWithNan = (writePosition - this.length) / 9;// TODO:
																			// stimmt
																			// Berechnung?
				for (int i = 0; i < rowsToFillWithNan; i++) {
					dos.writeDouble(Double.NaN); // TODO: festlegen welcher Wert
													// undefined sein soll NaN
													// ok?
					dos.writeByte(State.invalid.getCode()); // TODO: festlegen
															// welcher Wert
															// undefined sein
															// soll 00 ok?
					length += 9;
				}
				dos.writeDouble(value);
				dos.writeByte(flag);
				length += 9;
			}
		}
		/*
		 * close(); OutputStreams will not be closed or flushed. Data will be
		 * written to disk after calling flush() method.
		 */
	}

	public long getTimestampForLatestValue() {
		return startTimeStamp + (((this.length - 16) / 9) - 1) * storagePeriod;
	}

	/**
	 * calculates the position in a file for a certain timestamp
	 * 
	 * @param timestamp
	 * @return position
	 */
	private long getBytePosition(long timestamp) {
		if (timestamp >= startTimeStamp) {

			/*
			 * get position for timestamp 117 000: 117 000 - 100 000 = 17 000 17
			 * * 000 / 5 000 = 3.4 Math.round(3.4) = 3 3*(8+1) = 27 27 + 16 = 43
			 * = position to store to!
			 */
			// long pos = (Math.round((double) (timestamp - startTimeStamp) /
			// storagePeriod) * 9) + 16; /* slower */

			double pos = (double) (timestamp - startTimeStamp) / storagePeriod;
			if (pos % 1 != 0) { /* faster */
				pos = Math.round(pos);
			}
			return (long) (pos * 9 + 16);
		}
		else {
			// not in file! should never happen...
			return -1;
		}
	}

	/*
	 * Calculates the closest timestamp to wanted timestamp getByteposition does
	 * a similar thing (Math.round()), for byte position.
	 */
	private long getClosestTimestamp(long timestamp) {
		// return Math.round((double) (timestamp -
		// startTimeStamp)/storagePeriod)*storagePeriod+startTimeStamp; /*
		// slower */

		double ts = (double) (timestamp - startTimeStamp) / storagePeriod;
		if (ts % 1 != 0) {
			ts = Math.round(ts);
		}
		return (long) ts * storagePeriod + startTimeStamp;
	}

	public Value read(long timestamp) throws IOException {
		timestamp = getClosestTimestamp(timestamp); // round to: startTimestamp
		// + n*stepIntervall
		if (timestamp >= startTimeStamp && timestamp <= getTimestampForLatestValue()) {
			if (!canRead)
				enableInput();
			fis.getChannel().position(getBytePosition(timestamp));
			Double toReturn = dis.readDouble();
			if (!Double.isNaN(toReturn))
				return new Value(toReturn, timestamp, State.newState((int) dis.readByte()));
		}
		return null;
	}

	/**
	 * Returns a List of Value Objects containing the measured Values between
	 * provided start and end timestamp
	 * 
	 * @param start
	 * @param end
	 * @throws IOException
	 */
	public List<Value> read(long start, long end) throws IOException {
		start = getClosestTimestamp(start); // round to: startTimestamp +
											// n*stepIntervall
		end = getClosestTimestamp(end); // round to: startTimestamp +
										// n*stepIntervall

		List<Value> toReturn = new Vector<Value>();

		if (start < end) {
			if (start < startTimeStamp) // start & end should not exceed range
										// of this file.
				start = startTimeStamp;
			if (end > getTimestampForLatestValue())
				end = getTimestampForLatestValue();

			if (!canRead)
				enableInput();

			long timestampcounter = start;
			long startPos = getBytePosition(start);
			long endPos = getBytePosition(end);

			fis.getChannel().position(startPos);

			byte[] b = new byte[(int) (endPos - startPos) + 9];
			dis.read(b, 0, b.length);
			ByteBuffer bb = ByteBuffer.wrap(b);
			bb.rewind();

			for (int i = 0; i <= (endPos - startPos) / 9; i++) {
				double d = bb.getDouble();
				State s = State.newState((int) bb.get());
				if (!Double.isNaN(d))
					toReturn.add(new Value(d, timestampcounter, s));
				timestampcounter += storagePeriod;
			}

		}
		else if (start == end) {
			toReturn.add(read(start));
			toReturn.removeAll(Collections.singleton(null));
		}
		return toReturn; // Always return a list -> might be empty -> never is
							// null, to avoid NP's
	}

	public List<Value> readFully() throws IOException {
		return read(startTimeStamp, getTimestampForLatestValue());
	}

	/**
	 * Closes and Flushes underlying Input- and OutputStreams
	 * 
	 * @throws IOException
	 */
	public void close() throws IOException {
		canRead = false;
		canWrite = false;
		if (dos != null) {
			dos.flush();
			dos.close();
			dos = null;
		}
		if (fos != null) {
			fos.close();
			fos = null;
		}
		if (dis != null) {
			dis.close();
			dis = null;
		}
		if (fis != null) {
			fis.close();
			fis = null;
		}
	}

	/**
	 * Flushes the underlying Data Streams.
	 * 
	 * @throws IOException
	 */
	public void flush() throws IOException {
		if (dos != null)
			dos.flush();
	}
}
