package org.openmuc.mux.dbprovicer.slotsdb.test;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Simon Fey
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;

import org.openmuc.core.measurement.State;
import org.openmuc.core.measurement.Value;
import org.openmuc.mux.dbprovider.slotsdb.SlotsDatabase;
import org.openmuc.mux.dbprovider.slotsdb.FileObject;

import junit.framework.TestCase;

public class FileObjectTest extends TestCase {
	
	private FileObject testObject;
	private File testFile;
	
	public void setUp() throws Exception{
		super.setUp();
		File root = new File(SlotsDatabase.DB_TEST_ROOT_FOLDER);

		if(!root.exists())
			root.mkdirs();
		
		/*
		 * Deleting all Slots Database Files in JUnit Test folder
		 */
		for (File toDelete : root.listFiles()) {
			if (toDelete.isFile() && toDelete.getName().contains(SlotsDatabase.FILE_EXTENSION))
				toDelete.delete();
		}
		
		/*
		 * Create single FileObject for following tests.
		 */
		try {
			testObject = new FileObject(SlotsDatabase.DB_TEST_ROOT_FOLDER + "test" + SlotsDatabase.FILE_EXTENSION);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		/*
		 * create direct access to File
		 */
		testFile = new File(SlotsDatabase.DB_TEST_ROOT_FOLDER + "test" + SlotsDatabase.FILE_EXTENSION);
	}
	
	public void testCreateFileObject() {
		if (testFile.exists())
			fail("File should not suddenly exist after instantiating a FileObject.");
		
		try {
			testObject.createFileAndHeader(123456, 7000);
		} catch (IOException e) {
			e.printStackTrace();
		}
		if (!testFile.exists())
			fail("File should exist after calling createFile");
	}
	
	public void testOpenFileObject() {
		try {
			if(testFile.exists()) {
				testFile.delete();
			}
			testFile.createNewFile();
			
			DataOutputStream dos = new DataOutputStream(new FileOutputStream(testFile));
			dos.writeLong(1234567);
			dos.writeLong(8000);
			
			testObject = new FileObject(testFile);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		assertEquals(1234567, testObject.getStartTimeStamp());
		assertEquals(8000, testObject.getStoringPeriod());
	}
	
	public void testCreateAndAppend() throws IOException{
		FileObject fut = new FileObject(testFile);
		fut.createFileAndHeader(100000, 1000);
		fut.append(0.0, 100000, (byte)0x00);
		fut.append(0.1, 101000, (byte)0x00);
		/*
		 * leave out some:
		 * fut.append(0.2, 102000, (byte)0x00);
		 * fut.append(0.3, 103000, (byte)0x00);
		 */
		fut.append(0.4, 103600, (byte)0x00); //no exact timestamp, should be rounded to 10400
		fut.append(0.5, 105000, (byte)0x00); //no exact timestamp, should be rounded to 10500
		/*
		 * trying to re-add some:
		 * should be ignored!
		 */
		fut.append(0.8, 104000, (byte)0x00);
		fut.append(0.9, 105000, (byte)0x00);
		
		//SlotsDatabaseUtil.printWholeFile(testFile);
		fut.flush();
		
		if(testFile.length() != (16+(6*9)))
			fail("File does not have correct size!");
		
		DataInputStream dis = new DataInputStream(new FileInputStream(testFile));
		
		//Header
		assertEquals(100000, dis.readLong());
		assertEquals(1000, dis.readLong());
		
		assertEquals(0.0, dis.readDouble());
		assertEquals((byte)0x00, dis.readByte());
		assertEquals(0.1, dis.readDouble());
		assertEquals((byte)0x00, dis.readByte());
		
		assertEquals(Double.NaN, dis.readDouble());
		assertEquals(State.invalid.getCode(), dis.readByte());
		assertEquals(Double.NaN, dis.readDouble());
		assertEquals(State.invalid.getCode(), dis.readByte());
		
		assertEquals(0.4, dis.readDouble());
		assertEquals((byte)0x00, dis.readByte());
		assertEquals(0.5, dis.readDouble());
		assertEquals((byte)0x00, dis.readByte());

		if(dis.available() != 0)
			fail("File should be read fully.");
	}
	
	public void testReadSingleValue() throws IOException{
		DataOutputStream dos = new DataOutputStream(new FileOutputStream(testFile));
		/**
		 * Creating this testFile:
		 * 
		 * Start | Step
		 * 10000 | 1000
		 * 
		 * Value | Flag
		 * 0.9   | 0x01		(10000)
		 * 0.8   | 0x02		(11000)
		 * 0.7   | 0x03		(12000)
		 * 0.6   | 0x04		(13000)
		 * NaN   | 0x7F		(14000)
		 * NaN   | 0x7F		(15000)
		 * 0.3   | 0x05		(16000)
		 * 0.2   | 0x06		(17000)
		 */
		dos.writeLong(10000);
		dos.writeLong(1000);
		dos.writeDouble(0.9);
		dos.writeByte((byte) 0x01);
		dos.writeDouble(0.8);
		dos.writeByte((byte) 0x01);
		dos.writeDouble(0.7);
		dos.writeByte((byte) 0x00);
		dos.writeDouble(0.6);
		dos.writeByte((byte) 0x01);
		dos.writeDouble(Double.NaN);
		dos.writeByte((byte) Byte.MAX_VALUE);
		dos.writeDouble(Double.NaN);
		dos.writeByte((byte) Byte.MAX_VALUE);
		dos.writeDouble(0.3);
		dos.writeByte((byte) 0x01);
		dos.writeDouble(0.2);
		dos.writeByte((byte) 0x01);
		dos.close();
		
		FileObject testfo = new FileObject(testFile);
		assertEquals(10000, testfo.getStartTimeStamp());
		assertEquals(1000, testfo.getStoringPeriod());
		
		assertEquals(0.9, testfo.read(10200).getValue()); // ->10000
		assertEquals(State.valid, testfo.read(10000).getState()); // ->10000
		assertEquals(10000,testfo.read(10499).getTimestamp()); //->10000
		
		assertEquals(0.8, testfo.read(10500).getValue()); // ->11000
		assertEquals(State.valid, testfo.read(11000).getState());
		assertEquals(11000,testfo.read(11000).getTimestamp());
		
		assertEquals(0.7, testfo.read(12000).getValue());
		assertEquals(State.invalid, testfo.read(12000).getState());
		assertEquals(12000,testfo.read(12000).getTimestamp());
		
		assertEquals(0.6, testfo.read(13000).getValue());
		assertEquals(State.valid, testfo.read(13000).getState());
		assertEquals(13000,testfo.read(13000).getTimestamp());
		
		assertEquals(null, testfo.read(14000));
		
		assertEquals(null, testfo.read(15000));
		
		assertEquals(0.3, testfo.read(16000).getValue());
		assertEquals(State.valid, testfo.read(16000).getState());
		assertEquals(16000,testfo.read(16000).getTimestamp());
		
		assertEquals(0.2, testfo.read(17000).getValue());
		assertEquals(State.valid, testfo.read(17000).getState());
		assertEquals(17000,testfo.read(17000).getTimestamp());
	}
	
	public void testReadList() throws IOException {
		DataOutputStream dos = new DataOutputStream(new FileOutputStream(testFile));
		/**
		 * Creating this testFile:
		 * 
		 * Start | Step
		 * 10000 | 1000
		 * 
		 * Value | Flag
		 * 0.9   | 0x01		(10000)
		 * 0.8   | 0x02		(11000)
		 * 0.7   | 0x03		(12000)
		 * 0.6   | 0x04		(13000)
		 * NaN   | 0x7F		(14000)
		 * NaN   | 0x7F		(15000)
		 * 0.3   | 0x05		(16000)
		 * 0.2   | 0x06		(17000)
		 */
		dos.writeLong(10000);
		dos.writeLong(1000);
		dos.writeDouble(0.9);
		dos.writeByte((byte) 0x01);
		dos.writeDouble(0.8);
		dos.writeByte((byte) 0x01);
		dos.writeDouble(0.7);
		dos.writeByte((byte) 0x01);
		dos.writeDouble(0.6);
		dos.writeByte((byte) 0x01);
		dos.writeDouble(Double.NaN);
		dos.writeByte((byte) Byte.MAX_VALUE);
		dos.writeDouble(Double.NaN);
		dos.writeByte((byte) Byte.MAX_VALUE);
		dos.writeDouble(0.3);
		dos.writeByte((byte) 0x01);
		dos.writeDouble(0.2);
		dos.writeByte((byte) 0x01);
		dos.close();
		
		FileObject testfo = new FileObject(testFile);
		assertEquals(10000, testfo.getStartTimeStamp());
		assertEquals(1000, testfo.getStoringPeriod());
		
		List<Value> toCheck = testfo.read(10550, 30000); //will result in read from 11000 to 17000 -> 7 Values - 2 NaNs = 5 Values!
		assertEquals(5, toCheck.size()); //really are 5 Values?
		
		assertEquals(0.8, toCheck.get(0).getValue());
		assertEquals(11000, toCheck.get(0).getTimestamp());
		assertEquals(State.valid, toCheck.get(0).getState());
		
		assertEquals(0.7, toCheck.get(1).getValue());
		assertEquals(12000, toCheck.get(1).getTimestamp());
		assertEquals(State.valid, toCheck.get(1).getState());
		
		assertEquals(0.6, toCheck.get(2).getValue());
		assertEquals(13000, toCheck.get(2).getTimestamp());
		assertEquals(State.valid, toCheck.get(2).getState());
		
		assertEquals(0.3, toCheck.get(3).getValue());
		assertEquals(16000, toCheck.get(3).getTimestamp());
		assertEquals(State.valid, toCheck.get(3).getState());
		
		assertEquals(0.2, toCheck.get(4).getValue());
		assertEquals(17000, toCheck.get(4).getTimestamp());
		assertEquals(State.valid, toCheck.get(4).getState());
		testfo.close();
	}
	
	public void testReadWriteSwitch() throws IOException {
		DataOutputStream dos = new DataOutputStream(new FileOutputStream(testFile));
		dos.writeLong(10000);
		dos.writeLong(1000);
		
		FileObject fo = new FileObject(testFile.getPath());
		fo.append(1.0, 10000, (byte)0x01);
		assertEquals(1.0, fo.read(10000).getValue());
		fo.append(2.0, 11000, (byte)0x01);
		assertEquals(2.0, fo.read(11000).getValue());
		fo.append(3.0, 12000, (byte)0x01);
		fo.close();
		
		fo = new FileObject(testFile.getPath());
		assertEquals(3, fo.readFully().size());
	}
		
	public void tearDown() throws Exception{
		super.tearDown();
		setUp();
	}
}
