package org.openmuc.canopen.lib;

/*
 * Copyright Fraunhofer ISE, 2011
 * Author(s): Frederic Robra
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.LinkedList;
import java.util.List;

import org.openmuc.canopen.exception.NoDeviceException;
import org.openmuc.canopen.exception.TimeoutException;

import com.sun.jna.Native;

/**
 * Platform dependent library, needs driver compiled with netdev support.
 * Runs only on Linux.
 * Tested with PEAK microcontrol.<br>
 * <b>Protocol: CANopen</b>
 * @author Frederic Robra
 */
public class CANopenLib extends Thread {

	private int socket;

	private boolean running = false;
	
	public static long MS_TIMEOUT = 1000;
	
	List<SDOClient> sdoClients = new LinkedList<SDOClient>();
	
	private class Timeout extends Thread {
		private long msTime;
		private boolean end;
		public Timeout(long msTime) {
			this.msTime = msTime;
			this.end = false;
		}
		@Override
		public void run() {
			try {
				sleep(msTime);
			} catch (InterruptedException e) { }
			end = true;
			return;
		}
		public boolean isTimeout() {
			return end;
		}
	}
	
	public SDOClient newSDOClient(int nodeID) {
		SDOClient sdoClient = new SDOClient(nodeID, this);
		
		sdoClients.add(sdoClient);
		
		return sdoClient;
	}
	
	/**
	 * Creates a CANopenLib, with no connection.
	 */
	public CANopenLib(){
		socket = -1;
	}
	
	/**
	 * Connect to CAN
	 * @param ifname Name of the interface e.g. "can0"
	 * @throws NoDeviceException if interface isn't present.
	 * @throws IOException if connection can not be established.
	 */
	public void connect(String ifname) throws NoDeviceException, IOException{
		CLibrary.ifreq.ByReference ifr = new CLibrary.ifreq.ByReference();
		CLibrary.sockaddr_can.ByReference addr = new CLibrary.sockaddr_can.ByReference();
		/*
		 * Create a CAN socket
		 * C-Code:
		 * int skt = socket(PF_CAN, SOCK_RAW, CAN_RAW);
		 * if(skt < 0){
		 *   perror("Failed to create Socket");
		 *   return EXIT_FAILURE;
		 * }
		 */
		socket = CLibrary.INSTANCE.socket(CLibrary.PF_CAN, CLibrary.SOCK_RAW, CLibrary.CAN_RAW);
		if(socket < 0){
			throw new IOException("Failed to create Socket: " + nativeError());
		}
		CLibrary.INSTANCE.strcpy(ifr.ifr_ifrn.ifrn_name, ifname);
		/* IMPORTANT, if not present union will have no value in the native layer */
		ifr.ifr_ifrn.setType(ifr.ifr_ifrn.ifrn_name.getClass());
		/*
		 * Sets the name of the interface an returns the CAN interface index.
		 * C-Code:
		 * if(ioctl(skt, SIOCGIFINDEX, &ifr) < 0){
		 *   perror("IO-Control failed");
		 *   return EXIT_FAILURE;
		 * }
		 */
		if(CLibrary.INSTANCE.ioctl(socket, CLibrary.SIOCGIFINDEX, ifr) < 0){
			close();
			if(Native.getLastError() == CLibrary.ENODEV) {						/* If device does not exist */
				throw new NoDeviceException("IO-Control failed: " + nativeError());
			}
			else {
				throw new IOException("IO-Control failed: " + nativeError());
			}
		}
		
		addr.can_ifindex = ifr.ifr_ifru.ifru_ivalue;							/* Get interface index */
		addr.can_family = CLibrary.AF_CAN;
		/*
		 * Binds socket and address
		 * C-Code:
		 * if(bind(skt, (struct sockaddr *)&addr, sizeof(addr)) < 0){
		 *   perror("Bind failed");
		 *   return EXIT_FAILURE;
		 * }
		 */
		if(CLibrary.INSTANCE.bind(socket, addr, addr.size()) < 0){
			close();
			throw new IOException("Bind failed: " + nativeError());
		}
		
		/* Set socket to non-blocking mode, POSIX standard */
		int flag = CLibrary.INSTANCE.fcntl(socket, CLibrary.F_GETFL, 0);
		CLibrary.INSTANCE.fcntl(socket, CLibrary.F_SETFL, flag | CLibrary.O_NONBLOCK);
		
		
		running = true;
		this.start();
		/* TODO test write, network down */
	}
	
	@Override
	protected void finalize() throws Throwable {
		close();
		super.finalize();
	}
	
	/**
	 * Close CAN, method to close the socket.
	 * @param ifname Name of the Interface to close
	 */
	public void close(){
		running = false;
		
		//TODO wait until thread finished
		
		if(socket >= 0){
			CLibrary.INSTANCE.close(socket);
			socket = -1;
		}
	}
	
	
	private void printFrame(CLibrary.can_frame.ByReference frame, String prefix) {
		System.out.printf("%sID: %x DLC: %d ", prefix, frame.can_id, frame.can_dlc);
		for (int i = 0; i < frame.can_dlc; i++) {
			System.out.printf("%02x ", frame.data[i]);
		}
		System.out.println();
	}
	
	
	private void sendFrame(CLibrary.can_frame.ByReference frame) {
		printFrame(frame, "SEND ");
		if(CLibrary.INSTANCE.write(socket, frame, frame.size()) < 0) {
			System.err.println("write failed! " + nativeError());
		}
	}
	
	public void sendSDOUploadInitiate(byte node, int index, short subIndex) {
		CLibrary.can_frame.ByReference frame = new CLibrary.can_frame.ByReference();
		
		frame.can_id = 0x600 + node;
		
		frame.can_dlc = 8;
		
		frame.data[0] = 0x40;
		frame.data[1] = (byte) (index & 0x000000ff);
		frame.data[2] = (byte) (index >> 8);
		frame.data[3] = (byte) subIndex;
		frame.data[4] = 0;
		frame.data[5] = 0;
		frame.data[6] = 0;
		frame.data[7] = 0;
		
		sendFrame(frame);
	}
	
	public void sendSDOUploadSegment(byte node, boolean toggle) {
		CLibrary.can_frame.ByReference frame = new CLibrary.can_frame.ByReference();
		
		frame.can_id = 0x600 + node;
		
		frame.can_dlc = 8;
		
		if (toggle) 
			frame.data[0] = 0x70;
		else
			frame.data[0] = 0x60;
		
		frame.data[1] = 0;
		frame.data[2] = 0;
		frame.data[3] = 0;
		frame.data[4] = 0;
		frame.data[5] = 0;
		frame.data[6] = 0;
		frame.data[7] = 0;
		
		sendFrame(frame);
	}
	
	public
	CLibrary.can_frame.ByReference
	readFrameWithTimeout(int timeoutMs) {	

		CLibrary.can_frame.ByReference frame = new CLibrary.can_frame.ByReference();
		
		/* Set native select with timeout */
		CLibrary.fd_set.ByReference set = new CLibrary.fd_set.ByReference();
		CLibrary.timeval.ByReference timeout = new CLibrary.timeval.ByReference();
		timeout.tv_sec.setValue(0);
		timeout.tv_usec.setValue(timeoutMs * 1000);
		set.set(socket);
		
		/* Wait for some data */
		int select = CLibrary.INSTANCE.select(CLibrary.FD_SETSIZE, set, null, null, timeout);
		if(select <= 0)
			return null;
		
		/* Read data */
		int read = CLibrary.INSTANCE.read(socket, frame, frame.size());
		
		if(read < 0) {
			System.err.println("Read failed: " + nativeError());
			return null;
		}
		
		printFrame(frame, "RCVD ");

		return frame;
	}
	
	/**
	 * Sends a Message to the specific CAN.
	 * @param modulID ID of the Module, max=127.
	 * @param index The index of the Message, see CANopen.
	 * @param subindex The subindex of the Message, see CANopen.
	 * @return 4 Byte Data, either a String or a BigEndian value.
	 * @throws IOException if parameter have wrong format, or read returns wrong data, or no socket.
	 * @throws TimeoutException if read times out
	 */
	public byte[] getData(int modulID, int index, int subindex)	throws IOException, TimeoutException {
		if(socket < 0){
			throw new IOException("Send failed: Not connected to socket!");
		}
		
		byte[] cindex = new byte[2];
		cindex[0] = (byte) (index & 0xFF);
		index = index >> 8;
		cindex[1] = (byte) (index & 0xFF);
		
		CLibrary.can_frame.ByReference frame = new CLibrary.can_frame.ByReference();
		frame.can_id = 0x600 + modulID;
		frame.can_dlc = 0x08;
		frame.data[0] = 0x40;
		frame.data[1] = cindex[0];
		frame.data[2] = cindex[1];

		frame.data[3] = (byte) (subindex & 0xFF);

		frame.data[4] = 0x0;
		frame.data[5] = 0x0;
		frame.data[6] = 0x0;
		frame.data[7] = 0x0;
		
		if(CLibrary.INSTANCE.write(socket, frame, frame.size()) < 0) {
			throw new IOException("Write failed: " + nativeError());
		}
//		/*
//		 * Set a fd_set with socket and a timeout with 1 second.
//		 * Wait with select for timeout or file descriptor to receive something.
//		 */
//		CLibrary.fd_set.ByReference set = new CLibrary.fd_set.ByReference();
//		CLibrary.timeval.ByReference timeout = new CLibrary.timeval.ByReference();
//		timeout.tv_sec.setValue(1);
//		set.set(socket);
//		if(CLibrary.INSTANCE.select(CLibrary.FD_SETSIZE, set, null, null, timeout) <= 0){
//			throw new TimeoutException("Timeout: Could not read from modul " + modulID);
//		}
//		
//		/* while read would block in blocking mode, so no more data is available */
//		while(true){
//			int read = CLibrary.INSTANCE.read(socket, frame, frame.size());
//			if(read < 0){
//				if(Native.getLastError() == CLibrary.EWOULDBLOCK) {
//					Native.setLastError(0);
//					break;
//				}
//				else {
//					throw new IOException("Read failed: " + nativeError());
//				}
//			}
//		}
//		
//		if((frame.can_id != (0x580 + modulID))							/* Check for errors */
//				|| (frame.can_dlc != 0x08)
//				|| (frame.data[0] != 0x42)
//				|| (frame.data[1] != cindex[0])
//				|| (frame.data[2] != cindex[1])
//				|| (frame.data[3] != (byte) (subindex & 0xFF))){
//			/* TODO analyze error */
//			throw new IOException("Read failed: Received wrong data from modul " + modulID);
//		}
		
		Timeout counter = new Timeout(MS_TIMEOUT);
		counter.start();
		
		do {
			/* Set native select with timeout */
			CLibrary.fd_set.ByReference set = new CLibrary.fd_set.ByReference();
			CLibrary.timeval.ByReference timeout = new CLibrary.timeval.ByReference();
			timeout.tv_sec.setValue(MS_TIMEOUT / 1000);
			set.set(socket);
			/* Wait for some data */
			int select = CLibrary.INSTANCE.select(CLibrary.FD_SETSIZE, set, null, null, timeout);
			if(select <= 0) {
				throw new TimeoutException("Timeout: No data on CAN-Bus");
			}
			/* Read data */
			int read = CLibrary.INSTANCE.read(socket, frame, frame.size());
			if(read < 0) {
				throw new IOException("Read failed: " + nativeError());
			}
			if(counter.isTimeout()) {
				throw new TimeoutException("Timeout: Could not read from modul " + modulID);
			}
		} while((frame.can_id != (0x580 + modulID))							/* continue, if right data received */
				|| (frame.data[1] != cindex[0])
				|| (frame.data[2] != cindex[1])
				|| (frame.data[3] != (byte) (subindex & 0xFF)));
		
		if(frame.data[0] == (byte)0x80) {
			throw new IOException("Read failed: Received error message from modul " + modulID);
		}
		
		byte[] data = new byte[4];
		
		data[0] = frame.data[4];
		data[1] = frame.data[5];
		data[2] = frame.data[6];
		data[3] = frame.data[7];
		
		return data;
	}

	
	
	/**
	 * Changes byte[4] from BigEndian data into float.
	 * @param data Array of data bytes, BigEndian.
	 * @return Float value.
	 */
	public float bigEndianToFloat(byte[] data) {
		ByteBuffer b = ByteBuffer.wrap(data);
		b.order(ByteOrder.LITTLE_ENDIAN);
		return b.getFloat();
	}
	/**
	 * Changes byte[2] from BigEndian data into signed short.
	 * @param data Array of data bytes, BigEndian.
	 * @return Signed short value.
	 */
	public short bigEndianToInt16(byte[] data){
		ByteBuffer b = ByteBuffer.wrap(data);
		b.order(ByteOrder.LITTLE_ENDIAN);
		return b.getShort();
	}
	/**
	 * Changes byte[3] from BigEndian data into signed int. Last byte gets filled up.
	 * @param data Array of data bytes, BigEndian.
	 * @return Signed Integer value.
	 */
	public int bigEndianToInt24(byte[] data) {
		int value = bigEndianToInt32(data);
		if((data[2] & 0x80) > 0){							/* Fill up with 1s if 24 bit would be negative */
			value += 0xFF000000;
		}
		return value;
	}
	/**
	 * Changes byte[4] from BigEndian data into signed int.
	 * @param data Array of data bytes, BigEndian.
	 * @return Signed Integer value.
	 */
	public int bigEndianToInt32(byte[] data) {
		ByteBuffer b = ByteBuffer.wrap(data);
		b.order(ByteOrder.LITTLE_ENDIAN);
		return b.getInt();
	}
	
	/**
	 * Method to get the last native error.
	 * @return String containing the last native error.
	 * @see Native
	 */
	private String nativeError(){
		return CLibrary.INSTANCE.strerror(Native.getLastError());
	}

	@Override
	public void run() {
		CLibrary.can_frame.ByReference frame;
		
		while (running) {
			
			
			frame = readFrameWithTimeout(100);
			
			if (frame != null) {
			
				for (SDOClient client : sdoClients) {
					client.handleFrame(frame);
				}
			}
		}
	}



}
