package org.openmuc.canopen.modul;

/*
 * Copyright Fraunhofer ISE, 2011
 * Author(s): Frederic Robra
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.IOException;
import java.util.HashMap;

import org.openmuc.canopen.constants.DS301;
import org.openmuc.canopen.constants.DS404;
import org.openmuc.canopen.exception.TimeoutException;
import org.openmuc.canopen.lib.CANopenLib;

/**
 * Manages one module in a CAN network.<br>
 * <b>Protocol: DS-404</b>
 * @author Frederic Robra
 */
public class CANopenDS404Modul {
	int modulID;
	CANopenLib lib;
	HashMap<Integer, Channel> channels;
	
	private class Channel {
		public double decimalOffset;
		public int indexOffset;
	}
	
	/**
	 * Initialize a CAN Modul.
	 * @param modulID ID of the modul.
	 * @param lib CANopenLib to connect.
	 */
	public CANopenDS404Modul(int modulID, CANopenLib lib) {
		this.modulID = modulID;
		this.lib = lib;
		channels = new HashMap<Integer, Channel>();
	}

	/**
	 * Gets the value of the specific CAN
	 * @param channelID 
	 * @return Physical value of the CAN 
	 * @throws IOException if parameter have wrong format, or read returns wrong data, or no socket.
	 * @throws TimeoutException if read times out
	 */
	public double getValue(int channelID) throws IOException, TimeoutException {
		if(channels.get(channelID) == null) {
			Channel channel = new Channel();
			for(int x : DS404.X_OFFSETS) {						/* Search the right offset for the channel */
				try {
					lib.getData(modulID, DS404.X_ANALOG_INPUT_PROCESS_VALUE + x, channelID);
					channel.indexOffset = x;
					break;
				}
				catch(Exception e) { }
			}
			channel.decimalOffset =  Math.pow(10, -1 * getDecimalDigits(channelID));
			channels.put(channelID, channel);
		}
		Channel channel = channels.get(channelID);
		byte[] data = lib.getData(modulID, DS404.X_ANALOG_INPUT_PROCESS_VALUE + channel.indexOffset, channelID);
		switch(channel.indexOffset) {								/* Calculate data, corresponding to their type */
			case DS404.X_FLOAT_OFFSET:
				return lib.bigEndianToFloat(data);
			case DS404.X_INT16_OFFSET:
				return lib.bigEndianToInt16(data) * channel.decimalOffset;
			case DS404.X_INT24_OFFSET:
				return lib.bigEndianToInt24(data) * channel.decimalOffset;
			case DS404.X_INT32_OFFSET:
				return lib.bigEndianToInt32(data) * channel.decimalOffset;
			default:
				throw new IOException("Device not supported");
		}
	}
	
	/**
	 * The Name of the device e.g. "1.ti" for temperature sensor.
	 * @return The device name.
	 * @throws IOException if parameter have wrong format, or read returns wrong data, or no socket.
	 * @throws TimeoutException if read times out
	 */
	public String getName() throws IOException, TimeoutException{
		return new String(lib.getData(modulID, DS301.DEVICE_NAME, 0));
	}

	/**
	 * Gets the Decimal Digits of the specific CAN
	 * @param channelID 
	 * @return Integer of Digits. 
	 * @throws IOException if parameter have wrong format, or read returns wrong data, or no socket.
	 * @throws TimeoutException if read times out
	 */
	public int getDecimalDigits(int channelID) throws IOException, TimeoutException {
		byte[] data = new byte[4];
		int subindex = channelID;
		data = lib.getData(modulID, DS404.DECIMAL_DIGITS, subindex);
		return lib.bigEndianToInt16(data);
	}
	
	/**
	 * All available channels.
	 * @return The number of Channels available 1...199.
	 * @throws TimeoutException if read times out
	 */
	public int getChannels() throws TimeoutException {
		int channels = 1;
		try {
			channels = lib.bigEndianToInt32(lib.getData(modulID, DS404.SENSOR_TYPE, 0));
		}
		catch (IOException e) { }
		return channels;
	}

}
