package org.openmuc.ehz.driver;

import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import org.apache.log4j.Logger;
import org.openmuc.core.driver.DataReceivedListener;
import org.openmuc.core.driver.DeviceURI;
import org.openmuc.core.measurement.ChannelIdentifier;
import org.openmuc.core.measurement.Measurement;
import org.openmuc.core.measurement.Unit;
import org.openmuc.ehz.exception.NotFoundException;
import org.openmuc.ehz.iec62056_21.IECDevice;
import org.openmuc.ehz.sml.SMLDevice;

public class Connection extends Thread {
	private static Logger logger = Logger.getLogger(Connection.class);
	private BlockingQueue<MeterReadoutTask> readoutTaskQueue;
	
	private boolean running = false;
	private Device device;

	private class MeterReadoutTask {
		public DataReceivedListener callback;
		public Object callbackHandle;
		public List<ChannelIdentifier> registers;

		public MeterReadoutTask(DataReceivedListener callback, Object callbackHandle,
				List<ChannelIdentifier> registers, DeviceURI url) {
			this.callback = callback;
			this.callbackHandle = callbackHandle;
			this.registers = registers;
		}

	}
	

	public Connection(String iface, String configStr) throws NotFoundException, IOException {
		readoutTaskQueue = new LinkedBlockingQueue<MeterReadoutTask>();
		/* if IEC 62056-21 or SML */
		if (configStr.contains("IEC")) {
			device = new IECDevice(iface, configStr);
			logger.debug("Created IEC 62056-21 Device");
		}
		else {
			device = new SMLDevice(iface, configStr);
			logger.debug("Created SML Device");
		}
	}
	
	@Override
	public void run() {
		running = true;
		
		while(running) {
			try {
				if(!readoutTaskQueue.isEmpty()) {
					long timestamp = System.currentTimeMillis();
					
					MeterReadoutTask task = readoutTaskQueue.take();

					try {
						Hashtable<String, Double> values = device.getValues();
						List<Measurement> data = new ArrayList<Measurement>();
						
						for(ChannelIdentifier register : task.registers) {
							double value = values.get(register.getObjectLocator());
							Measurement measurement = new Measurement(register, value, Unit.UNKNOWN, timestamp);
							data.add(measurement);
						}
												
						task.callback.dataReceived(task.callbackHandle, data);

					}
					catch(ParseException e) {
						logger.error(e.getMessage());
						logger.error(e.getStackTrace().toString());
						task.callback.readoutError(task.callbackHandle, DataReceivedListener.INVALID_DATA);
					}
					catch(RuntimeException e) {
						logger.error(e.getMessage());
						logger.error(e.getStackTrace().toString());
						task.callback.readoutError(task.callbackHandle, DataReceivedListener.TIMEOUT);
					}
					catch (IOException e) {
						logger.error(e.getMessage());
						logger.error(e.getStackTrace().toString());
						task.callback.readoutError(task.callbackHandle, DataReceivedListener.INVALID_DATA);
					}

				}

				Thread.yield();
			}
			catch(InterruptedException e) { }
		}
	}

	public void addReadoutTask(DataReceivedListener callback, Object handle,
			List<ChannelIdentifier> registers, DeviceURI url) 
	{
		readoutTaskQueue.add(new MeterReadoutTask(callback, handle, registers, url));
	}
	
	public void close() {
		device.close();
		device = null;
		running = false;
	}
	
}
