package org.openmuc.mbus.driver;

import java.io.IOException;
import java.text.ParseException;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeoutException;

import org.apache.log4j.Logger;
import org.openmuc.core.driver.DeviceDriverInterface;
import org.openmuc.core.driver.DeviceURI;
import org.openmuc.core.driver.DataReceivedListener;
import org.openmuc.core.measurement.ChannelIdentifier;
import org.openmuc.core.measurement.Measurement;
import org.openmuc.core.measurement.Unit;
import org.openmuc.core.obis.OBISCode;
import org.openmuc.jmbus.MBusASAP;
import org.openmuc.jmbus.MBusToOBISTranslator;
import org.openmuc.jmbus.VariableDataBlock;
import org.openmuc.jmbus.VariableDataResponse;

public class MBusConnection implements Runnable {
	private static Logger logger = Logger.getLogger(MBusConnection.class);
	
	private static int instanceCount = 0;

	protected byte data[] = new byte[255];
	protected MBusASAP mBusASAP;

	private Thread thread;

	private BlockingQueue<MeterReadoutTask> readoutTaskQueue;

	private class MeterReadoutTask {
		public DataReceivedListener callback;
		public Object callbackHandle;
		public List<ChannelIdentifier> registers;
		public DeviceURI url;

		public MeterReadoutTask(DataReceivedListener callback, 
				Object callbackHandle,
				List<ChannelIdentifier> registers, DeviceURI url) 
		{
			this.callback = callback;
			this.callbackHandle = callbackHandle;
			this.registers = registers;
			this.url = url;
		}

	}

	public MBusConnection(String iface, String configStr) throws IOException {

		logger.info("Create new connection with interface " + iface);

		readoutTaskQueue = new LinkedBlockingQueue<MeterReadoutTask>();

		mBusASAP = new MBusASAP(iface, configStr);

		thread = new Thread(this, "mbus driver connection " + instanceCount);
		instanceCount++;
		thread.start();
	}

	public void run() {
		MeterReadoutTask curTask;

		while (true) {

			try {
				curTask = readoutTaskQueue.take();
			} catch (InterruptedException e) {
				logger.error(e.getMessage());
				return;
			}

			MBusToOBISTranslator translator = MBusToOBISTranslator.getInstance();

			VariableDataResponse resp = null;
			try {
				resp = mBusASAP.readMeter(curTask.url.getAddress());
			} catch (TimeoutException te) {
				logger.warn("MBUS timeout");
				curTask.callback.readoutError(curTask.callbackHandle, 
						DeviceDriverInterface.ERROR_TIMEOUT);
			} catch (IOException ioe) {
				logger.error(" : MBUS: I/O error " + ioe.getMessage());
				curTask.callback.readoutError(curTask.callbackHandle, 
						DeviceDriverInterface.ERROR_TIMEOUT);
			}
			
			if (resp == null) {
				continue;
			}

			Vector<Measurement> result = new Vector<Measurement>();

			for (VariableDataBlock vdb : resp.getVariableDataBlocks()) {

				try {
					vdb.parse();
				} catch (ParseException e) {
					continue;
				}

				long timestamp = System.currentTimeMillis();

				Measurement meas;

				for (ChannelIdentifier register : curTask.registers) {

					/*
					 * Check if object locator is obis or div/vif
					 */
					MBusObjectLocator ol = new MBusObjectLocator(register.getObjectLocator());

					if (ol.getType() == MBusObjectLocator.TYPE_MBUS) {
						if (Arrays.equals(vdb.getVIB(), ol.getVifs())) {
							if (Arrays.equals(vdb.getDIB(), ol.getDifs())) {

								Number value = (Number) vdb.getData();
								vdb.getScaler();

								meas = new Measurement(register, value.doubleValue() * Math.pow(10f, vdb.getScaler()),
										Unit.fromUnitCode(vdb.getUnit()), timestamp);
								result.add(meas);
								break;
							}
						}

					} else {
						Number value = (Number) vdb.getData();
						
						String obisCodeStr = translator.getOBISCode(resp.getDeviceType(), vdb);
						
						if (obisCodeStr != null) {
							
							OBISCode obisCode = new OBISCode(translator.getOBISCode(resp.getDeviceType(), vdb));

							if (register.getObjectLocator().equals(obisCode.toString())) {
								meas = new Measurement(register, value.doubleValue()  * Math.pow(10f, vdb.getScaler()), Unit.fromUnitCode(vdb.getUnit()), timestamp);
								result.add(meas);
							}
						} 
						
					} /* if (ol.getType() == MBusObjectLocatorTest.TYPE_MBUS) */

				} /* for (MeterDataIdentifier register : registers) */

			} /* for (VariableDataBlock vdb : resp.getVariableDataBlocks()) */

			curTask.callback.dataReceived(curTask.callbackHandle, result);

			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	} /* run() */

	public void addReadoutTask(DataReceivedListener callback, Object callbackHandle,
			List<ChannelIdentifier> registers, DeviceURI url) {
		readoutTaskQueue.add(new MeterReadoutTask(callback, callbackHandle, registers, url));
	}
	
	public void close() {
		mBusASAP.closeSerialPort();
	}

} /* Class MBusConnection */
