package org.openmuc.metering.driver.s7plc;

import org.com.libnodave.Connection;

/**
 * 
 * Parse an object locator for the Siemens S7 PLC driver
 * 
 * Object locator format:
 * 
 * DB<data block no>.<byte offset>.<data type>
 * 
 * examples
 * DB20.2.uint16 means data block 20, byte offset 2, variable type 16bit unsigned integer
 * 
 * DB10.10.float maps to a REAL variable
 * 
 * @author mzillgit
 *
 */
public class S7ObjectLocator {
	
	public static final byte TYPE_BIT = 0;
	public static final byte TYPE_INT8 = 1;
	public static final byte TYPE_UINT8 = 2;
	public static final byte TYPE_INT16 = 3;
	public static final byte TYPE_UINT16 = 4;
	public static final byte TYPE_INT32 = 5;
	public static final byte TYPE_UINT32 = 6;
	public static final byte TYPE_INT64 = 7;
	public static final byte TYPE_UINT64 = 8;
	public static final byte TYPE_FLOAT = 9;
	public static final byte TYPE_DOUBLE = 10;
	
	private int memoryArea;
	private int areaAddress;
	private int offset;
	private byte type;
	private byte bitPos;
	
	public S7ObjectLocator(String locator) throws MalformendObjectLocatorException {
		int areaAddressPos = 0;
		
		if (locator.startsWith("DB")) {
			this.memoryArea = Connection.AREA_DB;
			areaAddressPos = 2;
		}
		else {
			throw new MalformendObjectLocatorException("Unknown area");
		}
		
		int offsetPos = locator.indexOf('.');
		try {
			areaAddress = Integer.parseInt(locator.substring(areaAddressPos, offsetPos));
		} catch (NumberFormatException e) {
			throw new MalformendObjectLocatorException("NumberFormatException in areaAddress");
		}

		int typePos = locator.indexOf(':');
		try {
			offset = Integer.parseInt(locator.substring(offsetPos + 1, typePos));
		} catch (NumberFormatException e) {
			throw new MalformendObjectLocatorException("NumberFormatException in offset");
		}
		
		String typeStr = locator.substring(typePos + 1);
		
		if (typeStr.equals("float")) {
			this.type = TYPE_FLOAT;
		}
		else if (typeStr.equals("uint8")) {
			this.type = TYPE_UINT8;
		}
		else if (typeStr.equals("int8")) {
			this.type = TYPE_INT8;
		}
		else if (typeStr.equals("uint16")) {
			this.type = TYPE_UINT16;
		}
		else if (typeStr.equals("int16")) {
			this.type = TYPE_INT16;
		}
		else if (typeStr.equals("uint32")) {
			this.type = TYPE_UINT32;
		}
		else if (typeStr.equals("int32")) {
			this.type = TYPE_INT32;
		}
		else if (typeStr.equals("double")) {
			this.type = TYPE_DOUBLE;
		}
		else if (typeStr.startsWith("bit(")) {
			this.type = TYPE_BIT;
			String bitPos = locator.substring(typePos + 5, typePos + 6);
			try {
				this.bitPos = Byte.parseByte(bitPos);
			} catch (Exception e) {
				throw new MalformendObjectLocatorException("Invalid bit locator");
			}
		}
		else
			throw new MalformendObjectLocatorException("Unknown or missing type");
	}

	public int getMemoryArea() {
		return memoryArea;
	}

	public int getAreaAddress() {
		return areaAddress;
	}

	public int getOffset() {
		return offset;
	}

	public int getDataLength() {
		
		switch(this.type) {
		case TYPE_BIT:
		case TYPE_UINT8:
		case TYPE_INT8:
			return 1;
		case TYPE_UINT16:
		case TYPE_INT16:
			return 2;
		case TYPE_UINT32:
		case TYPE_INT32:
		case TYPE_FLOAT:
			return 4;
		case TYPE_UINT64:
		case TYPE_INT64:
		case TYPE_DOUBLE:
			return 8;
		default:
			return -1;
		}
	}
	
	public byte getType() {
		return type;
	}
	
	public byte getBitPos() {
		return this.bitPos;
	}
	
//	public static void main(String[] args) {
//		try {
//			S7ObjectLocator loc = new S7ObjectLocator("DB3.0:float");
//			System.out.println("Area DB? " + loc.getAreaAddress() + "." + loc.getOffset() + " type:" + loc.getType());
//			
//			loc = new S7ObjectLocator("DB3.123:uint16");
//			System.out.println("Area DB? " + loc.getAreaAddress() + "." + loc.getOffset() + " type:" + loc.getType());
//			
//			loc = new S7ObjectLocator("DB18.24:double");
//			System.out.println("Area DB? " + loc.getAreaAddress() + "." + loc.getOffset() + " type:" + loc.getType());
//		} catch (MalformendObjectLocatorException e) {
//			e.printStackTrace();
//		}
//	}
	
}
