package org.openmuc.metering.driver.s7plc;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.List;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.com.libnodave.Connection;
import org.com.libnodave.Interface;
import org.openmuc.core.driver.BusURI;
import org.openmuc.core.driver.ChannelInformation;
import org.openmuc.core.driver.DataReceivedListener;
import org.openmuc.core.driver.DeviceDriverInterface;
import org.openmuc.core.driver.DeviceURI;
import org.openmuc.core.driver.MethodNotImplementedException;
import org.openmuc.core.measurement.Measurement;
import org.openmuc.core.measurement.ChannelIdentifier;


/**
 * 
 * Implementation of MeterDriverInterface for Siemens S7 PLC
 * 
 * Meter URL format:
 * 
 * plcs7:10.0.0.7:102?slot=2;rack=0;mpi=2
 * 
 * 
 * 
 * Object locator format:
 * 
 * DB<data block no>.<byte offset>.<data type>
 * 
 * examples
 * DB20.2.uint16 means data block 20, byte offset 2, variable type 16bit unsigned integer
 * 
 * DB10.10.float maps to a REAL variable
 * 
 * @author mzillgit
 *
 */

public class S7PLCDriverImpl implements DeviceDriverInterface {
	
	Logger logger = Logger.getLogger(S7PLCDriverImpl.class);

	private int c = 1;
	
	private Vector<PLCConnection> connectionPool;
	
	protected class PLCConnection {
		Interface ifc;
		DeviceURI url;
		Connection con;
	}
	
	protected void removeConnection(PLCConnection con) {
		this.connectionPool.remove(con);
	}
	
	protected PLCConnection getConnection(DeviceURI url) {
		PLCConnection ret = null;
		
		/* Check pool if connection already exists */
		for (PLCConnection plccon : connectionPool) {
			if (plccon.url.equals(url)) {
				// System.out.println("plcs7: Found connection in pool");
				ret = plccon;
				break;
			}
		}
		
		/* Create new connection */
		if (ret == null) {
			Interface ifc = null;
			
			logger.info("Create new interface");
			
			if (url.getAddress() != null) {
				ifc = new Interface("IF" + c, url.getComPort(), Integer.parseInt(url.getAddress()));
				c++;
			}
			else {
				ifc = new Interface("IF" + c, url.getComPort(), 102);
				c++;
			}
			
			Connection con;
			
			ifc.setTimeout(5000000);
			
			int mpi = 2;
			int rack = 0;
			int slot = 2;
			
			if (url.getParameter("mpi") != null) {
				mpi = new Integer(url.getParameter("mpi"));
				System.out.println("mpi: " + mpi);
			}
			if (url.getParameter("rack") != null) {
				rack = new Integer(url.getParameter("rack"));
				System.out.println("rack: " + rack);
			}
			if (url.getParameter("slot") != null) {
				slot = new Integer(url.getParameter("slot"));
				System.out.println("slot: " + slot);
			}
			
			con = new Connection(ifc, mpi, rack, slot);

			if (con.connectPLC() == true) {
				logger.info("plcs7: connection established to " + url.toString());
				PLCConnection plccon = new PLCConnection();
				plccon.url = url;
				plccon.con = con;
				plccon.ifc = ifc;
				connectionPool.add(plccon);
				ret = plccon;
			}
			else logger.error("plcs7: Connect failed!!!");

		}
		
		return ret;
	} /* getConnction() */
	
	
	
	public S7PLCDriverImpl() {
		this.connectionPool = new Vector<PLCConnection>();
	}



	public void configureInterface(String iface, String configStr) 
		throws MethodNotImplementedException
	{
		throw new MethodNotImplementedException();
	}

	public String getDriverID() {
		return "plcs7";
	}

	public List<String> getDeviceList(String busURL) {
		return null;
	}

	public void getMeterValuesRequest(DataReceivedListener listener,
			int valueSelectionIndicator) {
		// TODO Auto-generated method stub

	}

	public List<String> getSupportedOperations() {
		// TODO Auto-generated method stub
		return null;
	}

	public void readDeviceAsync(DeviceURI url,
			List<ChannelIdentifier> registers,
			DataReceivedListener callback, Object handle) 
	{		
		Thread readoutThread = new Thread(new ReadoutThread(this, registers, callback, handle, url));
		readoutThread.start();
	}
	
	/**
	 * 
	 * Write a register on connected meter / device / i/o 
	 * 
	 * @param register - register to write to
	 * @param value    - value to write to register
	 * @throws IOException 
	 */
	public void writeChannel(DeviceURI url, ChannelIdentifier register, double value) 
		throws IOException
	{
		try {
			S7ObjectLocator locator = new S7ObjectLocator(register.getObjectLocator());
			
			PLCConnection con = this.getConnection(url);
			
			byte[] buf;
			
			switch (locator.getType()) {
			
			case S7ObjectLocator.TYPE_BIT:
				int intVal = (int) value;
				if (intVal == 0) {
					con.con.clrBit(locator.getMemoryArea(), locator.getAreaAddress(),
									locator.getOffset(), locator.getBitPos());
				}
				else {
					con.con.setBit(locator.getMemoryArea(), locator.getAreaAddress(),
							locator.getOffset(), locator.getBitPos());					
				}
				break;
			
			case S7ObjectLocator.TYPE_FLOAT:
				buf = new byte[4];
				ByteBuffer bbuf = ByteBuffer.wrap(buf);
				bbuf.order(ByteOrder.BIG_ENDIAN);
				bbuf.putFloat(0, (float) value);
				con.con.writeBytes(locator.getMemoryArea(), locator.getAreaAddress(), 
									locator.getOffset(), 4, buf);
				break;
				
			case S7ObjectLocator.TYPE_UINT8:
				buf = new byte[1];
				buf[0] = (byte) (((int) value) & 0xff);
				con.con.writeBytes(locator.getMemoryArea(), locator.getAreaAddress(), 
						locator.getOffset(), 1, buf);
			
				
			}
			
		} catch (MalformendObjectLocatorException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	} /* readWriteRegister() */

	public List<Measurement> readDeviceSync(DeviceURI url,
			List<ChannelIdentifier> registers) 
	{
		return null;
	}

	public List<DeviceURI> getDeviceList(BusURI url) throws MethodNotImplementedException, IOException {
		throw new MethodNotImplementedException();
	}

	public List<ChannelInformation> getDeviceChannels(DeviceURI url) throws MethodNotImplementedException, IOException {
		throw new MethodNotImplementedException();
	}
}
