package org.openmuc.sunnywebbox;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Simon Fey
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.IOException;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Vector;
import java.util.Map.Entry;

import org.apache.log4j.Logger;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.JSONValue;
import org.json.simple.parser.JSONParser;
import org.json.simple.parser.ParseException;
import org.openmuc.core.driver.ChannelInformation;
import org.openmuc.core.driver.DeviceURI;
import org.openmuc.core.measurement.ChannelIdentifier;
import org.openmuc.core.measurement.Measurement;
import org.openmuc.core.measurement.Unit;
import org.openmuc.core.obis.DLMSUnit;

public class SunnyWebboxDevice {
	private DeviceURI url;
	private SunnyWebboxConnection connection;
	private Map<String, Object> json_map;
	private JSONObject json_obj;
	private JSONParser json_parser;
	private JSONArray json_array;
	
	private static Logger logger = Logger.getLogger(SunnyWebboxDevice.class);

	public SunnyWebboxDevice(DeviceURI url) throws UnknownHostException, SocketException {
		this(url.getComPort());
		this.url = url;
	}

	public SunnyWebboxDevice(String comPort) throws UnknownHostException, SocketException {
		this.connection = new SunnyWebboxConnection(comPort);
		json_map = new LinkedHashMap();
		json_parser = new JSONParser();
	}
	
	public DeviceURI getUrl() {
		return url;
	}
	
	/**
	 * Returns a List of Connected Devices.
	 * @return
	 * @throws IOException
	 * @throws ParseException
	 */
	public List<String> getListOfConnectedDevices() throws IOException, ParseException {
		List<String> toReturn = new Vector<String>();
		
		json_map.put("version", "1.0");
		json_map.put("proc", "GetDevices");
		json_map.put("id", "1");
		json_map.put("format", "JSON");
		
		String response = connection.sendJSONRequest(JSONValue.toJSONString(json_map));
		if (response != null) {
			json_obj = (JSONObject) json_parser.parse(response);
			json_obj = (JSONObject) json_obj.get("result");
			json_array = (JSONArray) json_obj.get("devices");
			ListIterator<JSONObject> iter = json_array.listIterator();
			while (iter.hasNext()) {
				json_obj = (JSONObject) iter.next();
				toReturn.add((String) json_obj.get("key"));
			}
		}
		return toReturn;
	}
	
	/**
	 * Returns
	 * @param channels
	 * @return
	 * @throws MalformedObjectLocatorException
	 * @throws IOException
	 * @throws ParseException
	 */
	public Vector<Measurement> getMeasurements(List<ChannelIdentifier> channels) throws MalformedObjectLocatorException, IOException {
		long timestamp = System.currentTimeMillis();
		Vector<Measurement> measurements = new Vector<Measurement>();
		HashMap<String,HashMap<String,ChannelIdentifier>> sortedChannels = new HashMap<String,HashMap<String,ChannelIdentifier>>(); 
		HashMap<String, ChannelIdentifier> toPutIn;
		String unitString;
		
		/*
		 * Every Device which is connected to a SMA webbox has lots of subchannels.
		 * They are sorted here.
		 * object locators are:
		 * device1:subchannelA
		 * device1:subchannelB
		 * device1:subchannelN
		 * device2:subchannelA
		 * device1:subchannelX
		 * device2:subchannelB
		 * 
		 * After sorting, hashmap contains ONE list for every device, which contains matching subchannels!
		 */
		for (ChannelIdentifier channel : channels) {
			String locator = channel.getObjectLocator();
			SunnyWebboxObjectLocator obj_loc = new SunnyWebboxObjectLocator(locator);
			
			/* old version with obj_loc= device:channel;			
			toPutIn = sortedChannels.get(obj_loc.getDevice());
			if(toPutIn == null) {
				sortedChannels.put(obj_loc.getDevice(), new HashMap<String,ChannelIdentifier>());
				sortedChannels.get(obj_loc.getDevice()).put(obj_loc.getSubChannel(),channel);
			} else {
				toPutIn.put(obj_loc.getSubChannel(),channel);
			} */
			
			toPutIn = sortedChannels.get(url.getAddress());
			if(toPutIn == null) {
				sortedChannels.put(url.getAddress(), new HashMap<String,ChannelIdentifier>());
				sortedChannels.get(url.getAddress()).put(obj_loc.getSubChannel(),channel);
			} else {
				toPutIn.put(obj_loc.getSubChannel(),channel);
			}
		}
		
		/*
		 * For every deviceList in Hashmap a JSON request will be generated to read all subchannels
		 */
		Iterator<Entry<String, HashMap<String, ChannelIdentifier>>> deviceList = sortedChannels.entrySet().iterator();
		while (deviceList.hasNext()) {
			Entry<String, HashMap<String, ChannelIdentifier>> toProcess = deviceList.next();
			
			json_map.clear();
			json_map.put("version", "1.0");
			json_map.put("proc", "GetProcessData");
			json_map.put("id", "1");
			json_map.put("format", "JSON");
			
			JSONObject devices = new JSONObject();
			JSONArray devicesArray = new JSONArray();
			Map devicesArrayContent = new LinkedHashMap();
			
			devicesArrayContent.put("key", toProcess.getKey());
			JSONArray channelsContent = new JSONArray();
			
			Iterator<Entry<String, ChannelIdentifier>> subchannelList = toProcess.getValue().entrySet().iterator();
			while(subchannelList.hasNext()) {
				Entry<String, ChannelIdentifier> subchannelToProcess = subchannelList.next();
				String locator = subchannelToProcess.getValue().getObjectLocator();
				SunnyWebboxObjectLocator obj_loc = new SunnyWebboxObjectLocator(locator);
				channelsContent.add(obj_loc.getSubChannel());
			}
			
			devicesArrayContent.put("channels", channelsContent);
			devicesArray.add(devicesArrayContent);
			devices.put("devices", devicesArray);
			json_map.put("params", devices);
			
			String response = connection.sendJSONRequest(JSONValue.toJSONString(json_map));
			
			/*
			 * Now Response is received and parsing starts.
			 */
			
			try {
				json_obj = (JSONObject) json_parser.parse(response);
			} catch (ParseException e) {
				logger.error("Unable to parse JSON Response from Sunny Webbox.");
				return null;
			}
			json_obj = (JSONObject) json_obj.get("result");
			json_array = (JSONArray) json_obj.get("devices");
			json_obj = (JSONObject) json_array.get(0);
			json_array = (JSONArray)json_obj.get("channels");
			
			ListIterator iter = json_array.listIterator();
			while (iter.hasNext()) {
				unitString = null;
				json_obj = (JSONObject) iter.next();
				double readValue = Double.NaN;
				try {
					readValue = Double.parseDouble((String)json_obj.get("value"));
				} catch (NumberFormatException nfe) {
					readValue = Double.NaN;
				}
				
				if (json_obj.get("unit") != null && json_obj.get("unit") != "") {
					unitString = (String)json_obj.get("unit");
				} 
				if (json_obj.get("value") != null && json_obj.get("value") != "") {
					try {
					readValue = Double.parseDouble((String)json_obj.get("value"));
					} catch (NumberFormatException nfe) {
						/*never mind, will be NAN or null*/
						readValue = Double.NaN;
					}
				} else {
					readValue = Double.NaN;
				}
				
				if(!Double.isNaN(readValue)) {
					Measurement meas = getDLMSConformMeasurement(toProcess.getValue().get(json_obj.get("name")), unitString, readValue, timestamp);
					measurements.add(meas);
				}
			}
		}
		return measurements;
	} /*getchannels*/
	
	public List<ChannelInformation> getChannelsForDevice(String device) throws IOException, ParseException {
		List<ChannelInformation> toReturn = new Vector<ChannelInformation>();
		String channelName;
		String unitString = null;
		int dlmsUnitCode = 0;
		double curValue;
		
		json_map.clear();
		json_map.put("version", "1.0");
		json_map.put("proc", "GetProcessData");
		json_map.put("id", "1");
		json_map.put("format", "JSON");
		
		JSONObject devices = new JSONObject();
		JSONArray devicesArray = new JSONArray();
		Map devicesArrayContent = new LinkedHashMap();
		
		devicesArrayContent.put("key", device);
		
		devicesArray.add(devicesArrayContent);
		devices.put("devices", devicesArray);
		json_map.put("params", devices);
		
		String response = connection.sendJSONRequest(JSONValue.toJSONString(json_map));
		
		/*
		 * Now Response is received and de-JSON starts.
		 */
		
		json_obj = (JSONObject) json_parser.parse(response);
		json_obj = (JSONObject) json_obj.get("result");
		json_array = (JSONArray) json_obj.get("devices");
		json_obj = (JSONObject) json_array.get(0);
		json_array = (JSONArray)json_obj.get("channels");
		
		if(json_array != null) {
			ListIterator iter = json_array.listIterator();
			while (iter.hasNext()) {
				unitString = null;
				dlmsUnitCode = 0;
				json_obj = (JSONObject) iter.next();
				channelName = (String)json_obj.get("name");
				if (json_obj.get("unit") != null && json_obj.get("unit") != "") {
					unitString = (String)json_obj.get("unit");
				} 
				if (json_obj.get("value") != null && json_obj.get("value") != "") {
					try {
						curValue = Double.parseDouble((String)json_obj.get("value"));
					} catch (NumberFormatException nfe) {
						curValue = Double.NaN;
					}
				} else {
					curValue = Double.NaN;
				}
				
				if (unitString != null) {
					if (getDLMSUnitCode(unitString) == 0) {
						if (unitString.length()>1) {
							if(unitString.substring(0, 1).equalsIgnoreCase("k")) {
								if (getDLMSUnitCode(unitString.substring(1)) != 0){
									dlmsUnitCode = getDLMSUnitCode(unitString.substring(1));
								}
							} else if (unitString.substring(0, 1).equalsIgnoreCase("g")) {
								if (getDLMSUnitCode(unitString.substring(1)) != 0){
									dlmsUnitCode = getDLMSUnitCode(unitString.substring(1));
								}
							}
						}
					} else {
						dlmsUnitCode = getDLMSUnitCode(unitString);
					}
				}
				
				toReturn.add(new ChannelInformation(channelName, dlmsUnitCode, SunnyWebboxChannelDescription.getDescription(channelName)));
			}
		} else {
			logger.info("No channels found for device: "+device);
		}
		return toReturn;
	}
	
	/*
	 * gets the DLMS Unit Code from a string.
	 * If string starts with k: e.g. kWh -> unit Code for Wh is returned and value is multiplicated by 1000.
	 */
	private Measurement getDLMSConformMeasurement(ChannelIdentifier chanID, String unitString, double value, long timestamp) {
		int dlmsUnitCode = 0;
		double retVal = value;
		if (unitString != null) {
			if (getDLMSUnitCode(unitString) == 0) {
				if (unitString.length()>1) {
					if(unitString.substring(0, 1).equalsIgnoreCase("k")) {
						if (getDLMSUnitCode(unitString.substring(1)) != 0){
							dlmsUnitCode = getDLMSUnitCode(unitString.substring(1));
							retVal = 1000*value;
						}
					} else if (unitString.substring(0, 1).equalsIgnoreCase("g")) {
						if (getDLMSUnitCode(unitString.substring(1)) != 0){
							dlmsUnitCode = getDLMSUnitCode(unitString.substring(1));
							retVal = 1000000*value;
						}
					}
				}
			} else {
				dlmsUnitCode = getDLMSUnitCode(unitString);
			}
		}
		Measurement meas = new Measurement(chanID, retVal, Unit.fromUnitCode(dlmsUnitCode), timestamp);
		return meas; 
	}

	private int getDLMSUnitCode(String unitString) {
		if (unitString != null) {
			if(unitString.equals("%"))
				return DLMSUnit.PERCENTAGE;
			
			try {
				switch (unitCodes.valueOf(unitString)) {
				case Wh: return DLMSUnit.WATT_HOUR;
				case VAr: return DLMSUnit.VAR;
				case W: return DLMSUnit.WATT;
				case A: return DLMSUnit.AMPERE;
				case V: return DLMSUnit.VOLT;
				case Hz: return DLMSUnit.HERTZ;
				case h: return DLMSUnit.HOUR;
				case degC: return DLMSUnit.DEGREE_CELSIUS;
				case d: return DLMSUnit.DAY;
				default: return 0;
				} 
			} catch (IllegalArgumentException iae) {
				return 0;
			}
		}
		return 0;
	}
	
	enum unitCodes {
		Wh, VAr, W, A, V, Hz, h, degC, hhmmss, d
	}
}
