package org.openmuc.wago;

import java.util.List;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.openmuc.core.driver.DeviceURI;
import org.openmuc.core.measurement.Measurement;
import org.openmuc.core.measurement.ChannelIdentifier;
import org.openmuc.core.measurement.Unit;
import org.openmuc.core.measurement.Value;

import net.wimpi.modbus.ModbusException;
import net.wimpi.modbus.procimg.InputRegister;

public class Wago750Device {
	private static Logger logger = Logger.getLogger(Wago750Device.class);
	
	private Wago750DeviceNode[] nodes;

	private int analogInputRegisterCount = 0;
	private	int digitalInCount = 0;
	private int analogOutputRegisterCount = 0;
	private int digitalOutCount = 0;
	private int outRegisterCount = 0; /* Number of output registers in input process mapping */
	private int inRegisterCount = 0;  /* Number of input registers in input process mapping */
  	
	private WagoConnection connection;
	private DeviceURI url = null;
	
	public Wago750Device(DeviceURI url) throws Exception {
		this(url.getAddress());
		this.url = url;
	}
	
	public Wago750Device(String serverIP) throws Exception {

		this.connection = new WagoConnection(serverIP);
		
		try {
			nodes = getDeviceConfig();
		} catch (ModbusException e) {
			logger.error(e.getMessage());
		}
	}
	
	
	
	public DeviceURI getUrl() {
		return url;
	}



	public Wago750DeviceNode[] getNodes() {
		return this.nodes;
	}
	
	public void disconnect() {
		this.connection.close();
	}
	
	
	public void writeOutputChannel(ChannelIdentifier channel, double value) {
		try {
			WagoObjectLocator wol = new WagoObjectLocator(channel.getObjectLocator());
		
			Wago750DeviceNode node = nodes[wol.getModuleNo()];
			
			node.writeReg(wol.getChannel(), value);
			
			
		} catch (MalformedObjectLocatorException e) {
			logger.warn(e.getMessage());
		}		
	}
	
	public Vector<Measurement> getChannels(List<ChannelIdentifier> channels) 
		throws Exception  
	{
		long timestamp = System.currentTimeMillis();
		InputRegister[] inRegs = null;
		InputRegister[] outRegs = null;
		
		if (inRegisterCount > 0) {
			inRegs = this.connection.getInputRegisters(0, inRegisterCount);
		}
		
		if (outRegisterCount > 0) {
			outRegs = this.connection.getInputRegisters(0x200, outRegisterCount);
		}
		
		Wago750DeviceNode node;
		
		double val;
		
		Vector<Measurement> measurements = new Vector<Measurement>();
		
		for (ChannelIdentifier channel : channels) {
			String locator = channel.getObjectLocator();
			WagoObjectLocator wol = new WagoObjectLocator(locator);
	
			node = nodes[wol.getModuleNo()];
			
			if (node.digital) {
				if (!node.output) {
					if (inRegs != null) {
						
						int regAddr = analogInputRegisterCount;
						
						int bitCount = node.digitalAddress + wol.getChannel();
						
						if (wol.getChannel() >= 8)
							throw new MalformedObjectLocatorException(
									"Invalid channel!");
						
						regAddr += (bitCount / 16);
						
						int bitAddr = bitCount % 16;
						
						int digitalInput = inRegs[regAddr].getValue();

						int mask = 1 << bitAddr;

						if ((digitalInput & mask) == mask)
							val = 1.0;
						else
							val = 0.0;

						Value value = new Value(val, timestamp);

						Measurement meas = new Measurement(channel, value, Unit.UNKNOWN);
						measurements.add(meas);
					}
				} 
				else {
					if (outRegs != null) {
						int regAddr = analogOutputRegisterCount;
						
						int bitCount = node.digitalAddress + wol.getChannel();
						
						if (wol.getChannel() >= 8)
							throw new MalformedObjectLocatorException(
									"Invalid channel!");
						
						
						regAddr += (bitCount / 16);
						
						int bitAddr = bitCount % 16;
						
						int digitalOutput = outRegs[regAddr].getValue();
						
						int mask = 1 << bitAddr;

						if ((digitalOutput & mask) == mask)
							val = 1.0;
						else
							val = 0.0;

						Value value = new Value(val, timestamp);

						Measurement meas = new Measurement(channel, value, Unit.UNKNOWN);
						measurements.add(meas);
					}
				}
			}
			else {
				int rawValue;

				if (wol.getChannel() <= node.channels) {

					if (!node.output) {
						rawValue = inRegs[node.startAddress + wol.getChannel()]
								.getValue();
					} else {
						rawValue = outRegs[node.startAddress + wol.getChannel()]
								.getValue();
					}

					Value value = new Value((double) rawValue, timestamp);

					Measurement meas = new Measurement(channel, value, Unit.UNKNOWN);
					measurements.add(meas);
				} else
					throw new MalformedObjectLocatorException(
							"Invalid channel!");
			}
		}
		
		return measurements;
	}

	private Wago750DeviceNode[] getDeviceConfig() throws Exception {
		InputRegister[] regs;
		Wago750DeviceNode[] ret = null;
	
		analogInputRegisterCount = 0;
		analogOutputRegisterCount = 0;
		digitalInCount = 0;
		digitalOutCount = 0;
		
		int moduleCount = 0;

		regs = connection.getInputRegisters(0x2030, 65);
		
		for (; moduleCount < regs.length; moduleCount++) {
			if (regs[moduleCount].getValue() == 0)
				break;
		}

		ret = new Wago750DeviceNode[moduleCount];

		for (int i = 0; i < moduleCount; i++) {
			ret[i] = new Wago750DeviceNode();
			ret[i].con = this.connection;
			ret[i].type = regs[i].getValue();

			if ((regs[i].getValue() & 0x8000) == 0x8000) {
				ret[i].digital = true;

				int channels = (regs[i].getValue() & 0x7f00) >> 8;

				if ((regs[i].getValue() & 0x0001) == 0x0001) {
					ret[i].description = "Digital input, " + channels + " channels";
					ret[i].output = false;
					ret[i].channels = channels;
					ret[i].digitalAddress = digitalInCount;
					digitalInCount += channels;
				} else if ((regs[i].getValue() & 0x0002) == 0x0002) {
					ret[i].description = "Digital output, " + channels
							+ " channels";
					ret[i].output = true;
					ret[i].channels = channels;
					ret[i].digitalAddress = digitalOutCount;
					digitalOutCount += channels;
				} else {
					ret[i].description = "Digital unknown";
					break;
				}
			} else {
				ret[i].digital = false;

				switch (regs[i].getValue()) {
				case 342:
					ret[i].description = "Fieldbus Coupler";
					break;
				case 452:
					ret[i].description = "2 AI 12bit, 0-20 mA";
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;
				case 453:
					ret[i].description = "4 AI 12bit, 0-20 mA";
					ret[i].output = false;
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 4;
					ret[i].dataLength = 4;
					ret[i].channels = 4;
					break;	
				case 454:
					ret[i].description = "2 AI 12bit, 4-20 mA";
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;
				case 455:
					ret[i].description = "4 AI 12bit, 4-20 mA";
					ret[i].output = false;
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 4;
					ret[i].dataLength = 4;
					ret[i].channels = 4;
					break;
				case 457:
					ret[i].description = "4 AI 12bit, DC, +-10V, single-ended";
					ret[i].output = false;
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 4;
					ret[i].dataLength = 4;
					ret[i].channels = 4;
					break;	
				case 459:
					ret[i].description = "4 AI 12bit, DC, 0-10V, single-ended";
					ret[i].output = false;
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 4;
					ret[i].dataLength = 4;
					ret[i].channels = 4;
					break;
				case 460:
					ret[i].description = "4 Channel Analog Input Module for RTD's";
					ret[i].output = false;
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 4;
					ret[i].dataLength = 4;
					ret[i].channels = 4;
					break;					
				case 461:
					ret[i].description = "2 Channel Analog Input Module for RTD's";
					ret[i].output = false;
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;
				case 465:
					ret[i].description = "2 AI 12bit, 0-20 mA, single-ended";
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;
				case 466:
					ret[i].description = "2 AI 12bit, 4-20 mA, single-ended";
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;	
				case 467:
					ret[i].description = "2 AI 12bit, DC, 0-10V, single-ended";
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;

				case 468:
					ret[i].description = "4 AI 12bit, DC, 0-10V, single-ended";
					ret[i].output = false;
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 4;
					ret[i].dataLength = 4;
					ret[i].channels = 4;
					break;
				case 470:
					ret[i].description = "2 AI 12bit, 0-20 mA, single-ended";
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;
				case 472:
					ret[i].description = "2 AI 16bit, 0-20 mA, single-ended";
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;
				case 473:
					ret[i].description = "2 AI 12bit, 4-20 mA, single-ended";
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;
				case 474:
					ret[i].description = "2 AI 16bit, 4-20 mA, single-ended";
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;
				case 476:
					ret[i].description = "2 AI 16bit, DC, +-10V, single-ended";
					ret[i].output = false;
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;
				case 477:
					ret[i].description = "2 Channel, AC/DC 0-10 V, Diff.";
					ret[i].output = false;
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;
				case 478:
					ret[i].description = "2 AI 16bit, DC, 0-10V, single-ended";
					ret[i].output = false;
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;
				case 479:
					ret[i].description = "2 AI 13bit, DC, +-10V, differential";
					ret[i].output = false;
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;
				case 480:
					ret[i].description = "2 AI 13bit, 0-20 mA, differential";
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 2;
					ret[i].dataLength = 2;
					ret[i].channels = 2;
					break;
				case 553:
					ret[i].description = "4 AO 12bit, DC, 0-20mA";
					ret[i].output = true;
					ret[i].startAddress = analogOutputRegisterCount;
					analogOutputRegisterCount += 4;
					ret[i].dataLength = 4;
					ret[i].channels = 4;
					break;
				case 555:
					ret[i].description = "4 AO 12bit, DC, 4-20mA";
					ret[i].output = true;
					ret[i].startAddress = analogOutputRegisterCount;
					analogOutputRegisterCount += 4;
					ret[i].dataLength = 4;
					ret[i].channels = 4;
					break;
				case 557:
					ret[i].description = "4 AO 12bit, DC, +-10V";
					ret[i].output = true;
					ret[i].startAddress = analogOutputRegisterCount;
					analogOutputRegisterCount += 4;
					ret[i].dataLength = 4;
					ret[i].channels = 4;
					break;
				case 559:
					ret[i].description = "4 AO 12bit, DC, 0-10V";
					ret[i].output = true;
					ret[i].startAddress = analogOutputRegisterCount;
					analogOutputRegisterCount += 4;
					ret[i].dataLength = 4;
					ret[i].channels = 4;
					break;
				case 638:
					ret[i].description = "2 Channel Up/Down Counter / 500 Hz";
					ret[i].output = false;
					ret[i].startAddress = analogInputRegisterCount;
					analogInputRegisterCount += 4;
					ret[i].dataLength = 4;
					ret[i].channels = 2;
					break;
				case 842:
					ret[i].description = "Programmable Fieldbus Controller";
					break;
				default:
					ret[i].description = "unknown";
					break;
				}
			}
		}
		
		inRegisterCount = analogInputRegisterCount;
		inRegisterCount += (digitalInCount / 16);
		if (digitalInCount % 16 > 0) inRegisterCount ++;
		
		outRegisterCount = analogOutputRegisterCount;
		outRegisterCount += (digitalOutCount / 16);
		if (digitalOutCount % 16 > 0) outRegisterCount ++;
		
		/* Adjust address for digital I/Os */
		for (int i = 0; i < ret.length; i++) {
			if (ret[i].digital == true) {
				if (ret[i].output == false) {
					ret[i].startAddress = analogInputRegisterCount;
				}
			}
		}

		return ret;
	}
}
