package org.openmuc.mux.writer;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Calendar;
import java.util.Date;
import java.util.Dictionary;
import java.util.Formatter;
import java.util.List;
import java.util.Vector;

import org.openmuc.core.datamanager.Authentication;
import org.openmuc.core.datamanager.DataManager;
import org.openmuc.core.datamanager.DataManagerListener;
import org.openmuc.core.datamanager.DataStorage;
import org.openmuc.core.measurement.ChannelIdentifier;
import org.osgi.framework.BundleContext;

/**
 * 
 * @author mzillgit
 *
 * CurrentDataFileWriter can be configured by the following system properties:
 * 
 * <a>org.openmuc.core.datamanager.writer.currentdata.filename</a> name of file with current data
 * <a>org.openmuc.core.datamanager.writer.currentdata.interval</a> writing interval in milliseconds
 * 
 */

public class CurrentDataFileWriter extends Thread implements DataManagerListener {

	private DataManager dataManager;
	private List<String> directory;
	private Vector<StorageValuePair> data;
	private long writingInterval;
	private String curDataFilename;
	private boolean running = false;
	private boolean writing = false;
	private String dataFileDir = null; 
	
	private long currentTimestamp = 0;
	
	private DecimalFormat decimalFormat = null;
	
	DataManagerTracker storeTracker;
	
	private class Value {
		long timestamp;
		double value;
	}
	
	private class StorageValuePair {
		DataStorage storage;
		Value value;
	}
	
	public boolean isRunning() {
		return running;
	}

	public void setRunning(boolean running) {
		this.running = running;
	}

	public CurrentDataFileWriter(BundleContext bc, DataManagerTracker tracker, 
			Dictionary<?, ?> properties) 
	{
		storeTracker = tracker;

		updateDataConfig();
		
		try {
			configure(properties);
			writing = true;
		} catch (Exception e) {
			writing = false;
		}
	}
	
	private void updateDataConfig() {
		dataManager = storeTracker.getDataManager();
		
		directory = dataManager.getDeviceDirectory();
		
		data = new Vector<StorageValuePair>();
		
		Authentication auth = new Authentication("root", "context");
		
		StorageValuePair svp;
		
		for (String label : directory) {
			svp = new StorageValuePair();
			svp.storage = dataManager.getDataStorage(label, auth);
			svp.value = null;
			data.add(svp);
		}
	}
	
	private void configure(Dictionary<?, ?> properties) throws Exception  {
		String strInterval;
				
		String decimalFormatStr;
		
		
		dataFileDir = (String) properties.get("org.openmuc.mux.writer.data.directory");
		
		if (dataFileDir == null) dataFileDir = System.getProperty("org.openmuc.mux.writer.data.directory");
		
		if (dataFileDir != null) {
			if (!dataFileDir.endsWith("/")) {
				dataFileDir = dataFileDir + "/";
			}
		}
		else {
	    	System.out.println("Property " + "org.openmuc.mux.writer.currentdata.filename" + " undefined: won't run!");
	    	throw new Exception("Property org.openmuc.mux.writer.data.directory undefined!");
		}
		
		curDataFilename = (String) properties.get("org.openmuc.mux.writer.currentdata.filename");
		
	    if (curDataFilename == null) curDataFilename = System.getProperty("org.openmuc.mux.writer.currentdata.filename");
	    
	    if (curDataFilename == null) {
	    	System.out.println("Property " + "org.openmuc.mux.writer.currentdata.filename" + " undefined: won't run!");
	    	throw new Exception("Property org.openmuc.mux.writer.currentdata.filename undefined!");
	    }
	    
	    strInterval = (String) properties.get("org.openmuc.mux.writer.currentdata.interval");
		
	    if (strInterval == null) 
	    	strInterval = System.getProperty("org.openmuc.mux.writer.currentdata.interval");
	    
	    if (strInterval == null) {
	    	throw new Exception("Property org.openmuc.mux.writer.currentdata.interval undefined!");
	    }
	    
	    writingInterval = Long.parseLong(strInterval);
	    
	    decimalFormatStr = (String) properties.get("org.openmuc.mux.writer.format");
	    
	    if (decimalFormatStr == null)
	    	decimalFormatStr = System.getProperty("org.openmuc.mux.writer.format");
	    
	    if (decimalFormatStr != null) {
			decimalFormat = new DecimalFormat(decimalFormatStr);
			
			DecimalFormatSymbols dfs = decimalFormat.getDecimalFormatSymbols();
			
			dfs.setDecimalSeparator('.');
			
			decimalFormat.setDecimalFormatSymbols(dfs);
	    }
	    
	}
	
	public void run() {

		PrintStream ps;
		FileOutputStream fos;
		Calendar cal;

		String tmpFilename = this.curDataFilename + ".tmp";
		File tmpFile;

		running = true;

		while (running) {

			if (writing) {

				try {
					/* Open file */
					fos = new FileOutputStream(tmpFilename);
					fos.getChannel().lock();
					ps = new PrintStream(fos);

					/* Write current date */
					cal = Calendar.getInstance();
					int h, m, s;

					int year, month, day;

					year = cal.get(Calendar.YEAR) - 2000;
					month = cal.get(Calendar.MONTH) + 1;
					day = cal.get(Calendar.DAY_OF_MONTH);

					ps.printf("DATE[YYMMDD]: %02d%02d%02d ", year, month, day);

					h = cal.get(Calendar.HOUR_OF_DAY);
					m = cal.get(Calendar.MINUTE);
					s = cal.get(Calendar.SECOND);

					ps.printf("TIME[HHMMSS]: %02d%02d%02d ", h, m, s);

					ps.printf("TIME_DEC[s]: %d ", (cal.getTimeInMillis() / 1000));

					for (StorageValuePair svp : data) {

						ps.print(svp.storage.getDataId().getLabel() + ": ");

						if (svp.storage.getCurrentValue() != null) {
							ps.print(svp.storage.getCurrentValue().getValue() + " ");
						} else
							ps.print("null ");
					}

					ps.println();

					ps.close();
					fos.close();

					tmpFile = new File(tmpFilename);
					tmpFile.renameTo(new File(this.curDataFilename));
				} catch (Exception e1) {
					System.err.println(new Date().toString() + ": Caught exception:");
					e1.printStackTrace();
				}
			}

			try { Thread.sleep(this.writingInterval); } catch (InterruptedException e) {}

		}

	} /* run() */

	public void initialized() {
		updateDataConfig();
		writing = true;
	}

	public void updated() {
		updateDataConfig();
	}
		
	private synchronized void saveData() {		
		FileOutputStream fos;
		PrintStream ps;
		
		Calendar cal = Calendar.getInstance();
		cal.setTime(new Date(currentTimestamp));
		
		Formatter formatter = new Formatter();
		
		formatter.format("%s%02d%02d%02d.dat", dataFileDir, cal.get(Calendar.YEAR) - 2000, cal.get(Calendar.MONTH) + 1, cal.get(Calendar.DAY_OF_MONTH));
		
		String filename = formatter.toString();
			
		File file = new File(filename);
		
		if (!file.exists()) {
			file = createNewFileWithHeader(filename);		
		}
		
		try {
			fos = new FileOutputStream(filename, true);
			ps = new PrintStream(fos);
			
			/* Write timestamp date */

			int h, m, s;

			h = cal.get(Calendar.HOUR_OF_DAY);
			m = cal.get(Calendar.MINUTE);
			s = cal.get(Calendar.SECOND);

			ps.printf("%02d%02d%02d ", h, m, s);
			ps.print(currentTimestamp / 1000); ps.print(" ");
			
			for (StorageValuePair svp : data) {
				if (svp.value == null) {
					ps.print("-");
				}
				else {
					if (decimalFormat == null)
						ps.print(svp.value.value);
					else
						ps.print(decimalFormat.format(svp.value.value));
					
					svp.value = null;
				}
				ps.print(" ");
			}
			
			ps.println();
			fos.close();
			
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private File createNewFileWithHeader(String filename) {
		File file = new File(filename);
		FileOutputStream fos;
		PrintStream ps;
		
		try {
			file.createNewFile();
			fos = new FileOutputStream(filename);
			ps = new PrintStream(fos);
			ps.print("#TIME TIME_DEC ");

			for (StorageValuePair svp : data) {
				ps.print(svp.storage.getDataId().getLabel());
				ps.print(" ");
			}
			ps.println();
			ps.println("#");
			fos.close();
			return file;
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}

	public synchronized void storingDataEvent(String label, long timestamp,
			double value)
	{
		if (dataFileDir != null) {
			if (currentTimestamp > 0) {
				if (timestamp != currentTimestamp) {	
					/* save now */
					saveData();
								
					currentTimestamp = timestamp;
				}
			}
			else {
				currentTimestamp = timestamp;
			}
			
			for (StorageValuePair svp : data) {
				
				if (svp.storage.getDataId().getLabel() == label) {
					svp.value = new Value();
					svp.value.timestamp = timestamp;
					svp.value.value = value;
				}
			}
		}
	}

	public void stateChanged(ChannelIdentifier id, org.openmuc.core.measurement.State newState) {/*nop*/}

}
