package org.openmuc.mbuswebtool.internal;

/*
 * Copyright Fraunhofer ISE, 2010
 * Author(s): Michael Zillgith
 *    
 * This file is part of openMUC.
 * For more information visit http://www.openmuc.org
 * 
 * openMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * openMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with openMUC.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

import java.io.IOException;
import java.net.URL;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.concurrent.TimeoutException;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;
import org.openmuc.core.common.SerialPort;
import org.openmuc.core.datamanager.config.ChannelConfiguration;
import org.openmuc.core.datamanager.config.ConfigurationService;
import org.openmuc.core.datamanager.config.DeviceConfiguration;
import org.openmuc.core.measurement.Unit;
import org.openmuc.jmbus.MBusASAP;
import org.openmuc.jmbus.Util;
import org.openmuc.jmbus.VariableDataBlock;
import org.openmuc.jmbus.VariableDataResponse;
import org.openmuc.webinterface.AjaxView;
import org.openmuc.webinterface.PluginCategory;
import org.openmuc.webinterface.PluginContext;
import org.openmuc.webinterface.RedirectView;
import org.openmuc.webinterface.ResourceLoader;
import org.openmuc.webinterface.View;
import org.openmuc.webinterface.WebInterfacePlugin;
import org.osgi.framework.BundleContext;

public class MBusWebToolPlugin implements WebInterfacePlugin {

	private static Logger logger = Logger.getLogger(MBusWebToolPlugin.class);

	private BundleContext context;
	private ResourceLoader loader;
	private ConfigurationService configService;
	private List<RegisterInfo> registers = null;
	private MeterInfo meterInfo = null;
	private MBusScan mbusScan;

	private class RegisterLabelDescription {
		public RegisterLabelDescription(String label, String description, int unitCode) {
			this.label = label;
			this.description = description;
			this.unitCode = unitCode;
		}

		public String label;
		public String description;
		public int unitCode;
	}

	public MBusWebToolPlugin(BundleContext context, ConfigurationService configService) {
		this.context = context;
		this.configService = configService;
		loader = new ResourceLoader(context);
		mbusScan = new MBusScan();
	}

	public Hashtable<String, String> getResources() {
		Hashtable<String, String> resources = new Hashtable<String, String>();

		resources.put("css", "css");

		return resources;
	}

	public String getApplicationAlias() {
		return "mbuswebtool";
	}

	public String getApplicationName() {
		return "M-Bus Tool";
	}

	public String getApplicationDescription() {
		return "Tool to configure the M-Bus driver.";
	}

	private void configureDevice(HashMap<String, RegisterLabelDescription> registers) {
		String deviceName = meterInfo.getVendor() + "." + meterInfo.getMeterId();
		DeviceConfiguration newMeter = configService.createDevice(deviceName);
		newMeter.setDescription("autogenerated");
		newMeter.setMeterUrl(meterInfo.getComString());
		newMeter.setVendorID(meterInfo.getVendor());
		newMeter.setDeviceID(meterInfo.getMeterId());

		for (String difVif : registers.keySet()) {
			RegisterLabelDescription register = registers.get(difVif);

			ChannelConfiguration newRegister;

			if ((register.label != null) && (!register.label.equals(""))) {
				newRegister = newMeter.createChannel(register.label);
			}
			else {
				newRegister = newMeter.createChannel(deviceName + "." + difVif);
			}

			newRegister.setObjectLocator("mbus:" + difVif);
			newRegister.setStoringPeriod(60);
			newRegister.setStorageType(ChannelConfiguration.STORE_TYPE_MEAN);
			newRegister.setDescription(register.description);
			newRegister.setUnit(Unit.fromUnitCode(register.unitCode));

			Calendar cal = new GregorianCalendar();
			cal.set(Calendar.YEAR, 2010);
			cal.set(Calendar.MONTH, 0);
			cal.set(Calendar.DAY_OF_MONTH, 1);
			cal.set(Calendar.HOUR_OF_DAY, 0);
			cal.set(Calendar.MINUTE, 0);
			cal.set(Calendar.SECOND, 0);
			cal.set(Calendar.MILLISECOND, 0);
			newRegister.setStoringStartTime(cal.getTime());
		}

		configService.persist();
	}

	private List<RegisterInfo> readMbus(String device, String address) throws IOException, TimeoutException {
		List<RegisterInfo> registers = new ArrayList<RegisterInfo>();

		MBusASAP connection;

		connection = new MBusASAP(device, "2400-8N1");

		VariableDataResponse response;
		try {
			response = connection.readMeter(address);
		} catch (TimeoutException e) {
			throw e;
		} finally {
			connection.closeSerialPort();
		}

		String vendor = Util.vendorID(response.getManufacturerID());
		int primaryAddress = (0x00ff & response.getAddress());

		byte[] reverseMeterId = new byte[response.getMeterID().length];

		int j = response.getMeterID().length - 1;
		for (int i = 0; i < response.getMeterID().length; i++) {
			reverseMeterId[i] = response.getMeterID()[j--];
		}

		String meterId = Util.composeHexStringFromByteArray(reverseMeterId);
		String medium = "" + response.getDeviceType();

		switch (response.getDeviceType()) {
		case 0x02:
			medium = medium + " electricity";
			break;
		case 0x03:
			medium = medium + " gas";
			break;
		case 0x04:
			medium = medium + " heat";
			break;
		case 0x05:
			medium = medium + " steam";
			break;
		case 0x06:
			medium = medium + " warm water";
			break;
		case 0x07:
			medium = medium + " water";
			break;
		case 0x16:
			medium = medium + " cold water";
			break;
		}

		String comString = "mbus:" + device + ":" + address;

		meterInfo = new MeterInfo(vendor, meterId, medium, primaryAddress, comString);

		List<VariableDataBlock> blocks = response.getVariableDataBlocks();

		for (VariableDataBlock dataBlock : blocks) {
			String difVif = Util.composeHexStringFromByteArray(dataBlock.getDIB()) + "/"
					+ Util.composeHexStringFromByteArray(dataBlock.getVIB());

			try {
				dataBlock.parse();
				if (dataBlock.getDescription() != null) {
					String description = dataBlock.getDescription().toString() + " "
							+ dataBlock.getFunctionField().toString();

					String value = dataBlock.getData().toString();
					String unit = Unit.getString(Unit.fromUnitCode(dataBlock.getUnit()));
					int unitCode = dataBlock.getUnit();
					registers.add(new RegisterInfo(difVif, description, value, unit, unitCode));
				}
			} catch (ParseException e) {
				registers.add(new RegisterInfo(difVif, "Unknown Channel", "", "", 0));
			}

		}

		connection.closeSerialPort();

		return registers;
	}

	public View getContentView(HttpServletRequest request, PluginContext context) {
		String template = "";

		if (context.getLocalPath().equals("/scanmeter")) {
			String device = request.getParameter("serialdevice");
			System.out.println(device);
			mbusScan = null;
			mbusScan = new MBusScan();
			mbusScan.setDevice(device);
			mbusScan.start();
			return new RedirectView(context.getApplicationPath());
		}
		else if (context.getLocalPath().equals("/scanstatus")) {
			template = loader.getResourceAsString("scanstatus.html");
			AjaxView view = new AjaxView(template);
			try {
				view.addToContext("primaryAddresses", mbusScan.getPrimaryAddresses());
				view.addToContext("progress", mbusScan.getProgress());
				view.addToContext("scanRunning", mbusScan.isRunning());
				view.addToContext("port", mbusScan.getDevice());
				if (!mbusScan.isRunning()) {
					view.addToContext("portList", SerialPort.getPortIdList());
				}
			} catch (Exception e) {
			}
			return view;
		}
		if (context.getLocalPath().equals("/scanstop")) {
			mbusScan.stopScan();
			try {
				mbusScan.join();
			} catch (InterruptedException e) {
			}
			return new RedirectView(context.getApplicationPath());
		}
		else if (context.getLocalPath().equals("/readmeter")) {
			mbusScan.stopScan();
			try {
				mbusScan.join();
			} catch (InterruptedException e) {
			}
			template = loader.getResourceAsString("message.html");
			MBusView view = new MBusView(template);
			String device = request.getParameter("serialdevice");
			// String config = request.getParameter("serialconfig");
			String mbusAddress = request.getParameter("mbusaddr");

			String message;

			try {
				registers = readMbus(device, mbusAddress);
				message = "OK!";
				template = loader.getResourceAsString("modulesview.html");
				view = new MBusView(template);
				view.addToContext("mbusMeter", meterInfo);
				view.addToContext("mbusRegisters", registers);

			} catch (IOException e) {
				logger.error(e.getMessage());
				message = "IO error!";
			} catch (TimeoutException e) {
				message = "Timeout!";
			} catch (Exception e) {
				message = "Error";
			}

			view.addToContext("mbusMessage", message);
			return view;
		}
		else if (context.getLocalPath().equals("/addmeter")) {
			template = loader.getResourceAsString("message.html");
			MBusView view = new MBusView(template);
			try {
				HashMap<String, RegisterLabelDescription> registers = new HashMap<String, RegisterLabelDescription>();

				String[] params = request.getParameterValues("registers");
				for (String param : params) {

					RegisterLabelDescription register = new RegisterLabelDescription(request.getParameter(param
							+ "label"), request.getParameter(param + "description"), new Integer(
							request.getParameter(param + "unitCode")));

					registers.put(param, register);
				}

				configureDevice(registers);

				view.addToContext("mbusMessage", "Configured  device");
			} catch (Exception e) {
				view.addToContext("mbusMessage", "Error configuring device");
			}
			return view;
		}
		else {
			template = loader.getResourceAsString("enterMeterInfo.html");
			List<String> portList = SerialPort.getPortIdList();

			MBusView view = new MBusView(template);
			view.addToContext("portList", portList);
			view.addToContext("scanRunning", mbusScan.isRunning());

			return view;
		}
	}

	public boolean handleSecurity(HttpServletRequest request, HttpServletResponse response) throws IOException {
		return true;
	}

	public URL getResource(String name) {
		return context.getBundle().getResource(name);
	}

	public String getMimeType(String name) {
		return null;
	}

	public PluginCategory getCategory() {
		return PluginCategory.TOOL;
	}

}
